"""A Markdown Component for details/information about DataSources (and FeatureSets)"""

from typing import Union
import logging

# Dash Imports
from dash import html, dcc

# Workbench Imports
from workbench.api import DataSource, FeatureSet
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.utils.markdown_utils import tags_to_markdown

# Get the Workbench logger
log = logging.getLogger("workbench")


class DataDetails(PluginInterface):
    """DataSource/FeatureSet Details Component"""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.NONE
    plugin_input_type = PluginInputType.DATA_SOURCE

    def __init__(self):
        """Initialize the DataDetails plugin class"""
        self.component_id = None
        self.data_source = None

        # Call the parent class constructor
        super().__init__()

    def create_component(self, component_id: str) -> html.Div:
        """Create a Markdown Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            html.Div: A Container of Components for the Model Details
        """
        self.component_id = component_id
        container = html.Div(
            id=self.component_id,
            children=[
                html.H3(id=f"{self.component_id}-header", children="Loading..."),
                dcc.Markdown(id=f"{self.component_id}-details", dangerously_allow_html=True),
            ],
        )

        # Fill in plugin properties
        self.properties = [
            (f"{self.component_id}-header", "children"),
            (f"{self.component_id}-details", "children"),
        ]

        # Return the container
        return container

    def update_properties(self, artifact: Union[DataSource, FeatureSet], **kwargs) -> list:
        """Update the properties for the plugin.

        Args:
            artifact (Union[DataSource, FeatureSet]): An instantiated DataSource or FeatureSet
            **kwargs: Additional keyword arguments (unused)

        Returns:
            list: A list of the updated property values for the plugin
        """
        log.important(f"Updating Plugin with Artifact: {artifact.name} and kwargs: {kwargs}")

        # Update the header and the details
        self.data_source = artifact.data_source if isinstance(artifact, FeatureSet) else artifact
        header = f"{self.data_source.name}"
        data_details_markdown = self.data_details_markdown(self.data_source.details())

        return [header, data_details_markdown]

    def data_details_markdown(self, data_details: dict) -> str:

        markdown_template = (
            "**Rows:** <<num_rows>>  \n"
            "**Columns:** <<num_columns>>  \n"
            "**Created/Mod:** <<created>> / <<modified>>  \n"
            "<<workbench_tags>>  \n"
            "**Input:** <<input>>  \n"
            "<br>  \n"
            "\n#### Numeric Columns  \n"
            "<<numeric_column_details>>  \n"
            "\n<div style='margin-top: 15px;'></div>  \n"
            "\n#### Non-Numeric Columns  \n"
            "<<string_column_details>>  \n"
        )

        expanding_list = (
            "<details>  \n"
            "    <summary><<column_info>></summary>  \n"
            "    <ul>  \n"
            "    <<bullet_list>>  \n"
            "    </ul>  \n"
            "</details>"
        )

        # Sanity Check for empty data
        if not data_details:
            return "No data source details found"

        # Loop through all the details and replace in the template
        for key, value in data_details.items():
            # Hack for dates
            if ".000Z" in str(value):
                try:
                    value = value.replace(".000Z", "").replace("T", " ")
                except AttributeError:
                    pass

            # Special case for tags
            if key == "workbench_tags":
                value = tags_to_markdown(value)
            markdown_template = markdown_template.replace(f"<<{key}>>", str(value))

        # Fill in numeric column details
        column_stats = data_details.get("column_stats", {})
        numeric_column_details = ""
        numeric_types = [
            "tinyint",
            "smallint",
            "int",
            "bigint",
            "float",
            "double",
            "decimal",
        ]
        for column_name, column_info in column_stats.items():
            if column_info["dtype"] in numeric_types:
                column_html = self._column_info_html(column_name, column_info)
                column_details = expanding_list.replace("<<column_info>>", column_html)

                # Populate the bullet list (descriptive_stats and unique)
                bullet_list = ""
                for q, value in column_info["descriptive_stats"].items():
                    if value is not None:
                        bullet_list += f"<li>{q}: {value:.3f}</li>"
                bullet_list += f"<li>Unique: {column_info['unique']}</li>"

                # Add correlations if they exist
                if column_info.get("correlations"):
                    corr_title = """<span class="green-text"><b>Correlated Columns</b></span>"""
                    corr_details = expanding_list.replace("<<column_info>>", corr_title)
                    corr_details = f"""<li class="no-bullet">{corr_details}</li>"""
                    corr_list = ""
                    for col, corr in column_info["correlations"].items():
                        corr_list += f"<li>{col}: {corr:.3f}</li>"
                    corr_details = corr_details.replace("<<bullet_list>>", corr_list)
                    bullet_list += corr_details

                # Add the bullet list to the column details
                column_details = column_details.replace("<<bullet_list>>", bullet_list)

                # Add the column details to the markdown
                numeric_column_details += column_details

        # Now actually replace the column details in the markdown
        markdown_template = markdown_template.replace("<<numeric_column_details>>", numeric_column_details)

        # For string columns create collapsible sections that show value counts
        string_column_details = ""
        for column_name, column_info in column_stats.items():
            # Skipping any columns that are numeric
            if column_info["dtype"] in numeric_types:
                continue

            # Skipping any columns that are dates/timestamps
            if column_info["dtype"] == "timestamp":
                continue

            # Create the column info
            column_html = self._column_info_html(column_name, column_info)
            column_details = expanding_list.replace("<<column_info>>", column_html)

            # Populate the bullet list (if we have value counts)
            if "value_counts" not in column_info:
                bullet_list = "<li>No Value Counts</li>"
            else:
                bullet_list = ""
                for value, count in column_info["value_counts"].items():
                    bullet_list += f"<li>{value}: {count}</li>"

            # Add the bullet list to the column details
            column_details = column_details.replace("<<bullet_list>>", bullet_list)

            # Add the column details to the markdown
            string_column_details += column_details

        # Now actually replace the column details in the markdown
        markdown = markdown_template.replace("<<string_column_details>>", string_column_details)
        return markdown

    @staticmethod
    def _construct_full_type(column_info: dict) -> dict:
        """Internal: Show the FeatureSet Types if they exist"""
        shorten_map = {
            "Integral": "I",
            "Fractional": "F",
            "String": "S",
            "Timestamp": "TS",
            "Boolean": "B",
        }
        if "fs_dtype" in column_info:
            display_fs_type = shorten_map.get(column_info["fs_dtype"], "V")
            column_info["full_type"] = f"{display_fs_type}: {column_info['dtype']}"
        else:
            column_info["full_type"] = column_info["dtype"]
        return column_info

    def _column_info_html(self, column_name, column_info: dict) -> str:
        """Internal: Create an HTML string for a column's information
        Args:
            column_name (str): The name of the column
            column_info (dict): A dictionary of column information
        Returns:
            str: An HTML string
        """

        # First part of the HTML template is the same for all columns
        html_template = """<b><<name>></b> <span class="blue-text">(<<full_type>>)</span>:"""

        # Add min, max, and number of zeros for numeric columns
        numeric_types = [
            "tinyint",
            "smallint",
            "int",
            "bigint",
            "float",
            "double",
            "decimal",
        ]
        float_types = ["float", "double", "decimal"]
        if column_info["dtype"] in numeric_types:
            # Just hardcode the min and max for now
            min = column_info["descriptive_stats"]["min"]
            max = column_info["descriptive_stats"]["max"]

            # Sanity Check
            if min is None or max is None:
                html_template += """ <span class="red-text">No Stats</span>"""

            # Floats get 2 decimal places
            elif column_info["dtype"] in float_types:
                html_template += f""" {min:.2f} → {max:.2f}&nbsp;&nbsp;&nbsp;&nbsp;"""

            # Integers get no decimal places
            else:
                html_template += f""" {int(min)} → {int(max)}&nbsp;&nbsp;&nbsp;&nbsp;"""
            if column_info["unique"] == 2 and min == 0 and max == 1:
                html_template += """ <span class="green-text"> Binary</span>"""
            elif column_info["num_zeros"] > 0:
                html_template += """ <span class="orange-text"> Zero: <<num_zeros>></span>"""

        # Non-numeric columns get the number of unique values
        else:
            html_template += """ Unique: <<unique>> """

        # Do we have any nulls in this column?
        if column_info["nulls"] > 0:
            html_template += """ <span class="red-text">Null: <<nulls>></span>"""

        # Replace the column name
        html_template = html_template.replace("<<name>>", column_name)

        # Construct the full type
        column_info = self._construct_full_type(column_info)

        # Loop through all the details and replace in the template
        for key, value in column_info.items():
            html_template = html_template.replace(f"<<{key}>>", str(value))

        return html_template


if __name__ == "__main__":
    # This class takes in model details and generates a details Markdown component
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

    # Run the Unit Test on the Plugin
    PluginUnitTest(DataDetails).run()



import IPython
from IPython import start_ipython
from distutils.version import LooseVersion
from IPython.terminal.prompts import Prompts
from IPython.terminal.ipapp import load_default_config
from pygments.token import Token
import sys
import logging
import importlib
import botocore
import webbrowser
import pandas as pd
import readline  # noqa

try:
    import matplotlib.pyplot as plt  # noqa

    plt.ion()
    HAVE_MATPLOTLIB = True
except ImportError:
    HAVE_MATPLOTLIB = False

try:
    import plotly.io as pio  # noqa

    pio.renderers.default = "browser"
    HAVE_PLOTLY = True
except ImportError:
    HAVE_PLOTLY = False


# Workbench Imports
from workbench.utils.repl_utils import cprint, Spinner
from workbench.utils.workbench_logging import IMPORTANT_LEVEL_NUM, TRACE_LEVEL_NUM
from workbench.utils.config_manager import ConfigManager
from workbench.utils.log_utils import silence_logs, log_theme
from workbench.api import Meta
from workbench.cached.cached_meta import CachedMeta

# If we have RDKIT/Mordred let's pull in our cheminformatics utils
try:
    import rdkit  # noqa
    import mordred  # noqa
    from workbench.utils import chem_utils

    HAVE_CHEM_UTILS = True
except ImportError:
    HAVE_CHEM_UTILS = False


def onboard():
    """Onboard a new user to Workbench"""
    cprint("lightgreen", "Welcome to Workbench!")
    cprint("lightblue", "Looks like this is your first time using Workbench...")
    cprint("lightblue", "Let's get you set up...")

    # Create a Site Specific Config File
    cm = ConfigManager()
    cm.create_site_config()
    cm.platform_specific_instructions()

    # Tell the user to restart the shell
    cprint("lightblue", "After doing these instructions ^")
    cprint("lightblue", "Please rerun the Workbench REPL to complete the onboarding process.")
    cprint("darkyellow", "Note: You'll need to start a NEW terminal to inherit the new ENV vars.")
    sys.exit(0)


# Check config and onboard if necessary
if not ConfigManager().config_okay():
    onboard()

# Set the log level to important
log = logging.getLogger("workbench")
log.setLevel(logging.INFO)
log.addFilter(
    lambda record: not (
        record.getMessage().startswith("Async: Metadata") or record.getMessage().startswith("Updated Metadata")
    )
)


# We want to customize our prompt colors
prompt_styles = {
    Token.Workbench: "#af87ff",  # Light Purple color for Workbench
    Token.AWSProfile: "#ffd700",  # Yellow color for AWS Profile
    Token.Lightblue: "#5fd7ff",
    Token.Lightpurple: "#af87ff",
    Token.Lightgreen: "#87ff87",
    Token.Lime: "#afff00",
    Token.Darkyellow: "#ddb777",
    Token.Orange: "#ff8700",
    Token.Red: "#dd0000",
    Token.Blue: "#4444d7",
    Token.Green: "#22cc22",
    Token.Yellow: "#ffd787",
    Token.Grey: "#aaaaaa",
}


# Note: Hack so the Prompt Class can access these variables
aws_profile = ConfigManager().get_config("AWS_PROFILE")
workbench_shell = None


class WorkbenchPrompt(Prompts):
    """Custom Workbench Prompt"""

    def in_prompt_tokens(self, cli=None):
        if workbench_shell is None:
            lights = []
        else:
            lights = workbench_shell.status_lights()
        aws_profile_prompt = [(Token.Blue, ":"), (Token.AWSProfile, f"{aws_profile}"), (Token.Blue, "> ")]
        return lights + [(Token.Workbench, "Workbench")] + aws_profile_prompt


class WorkbenchShell:
    def __init__(self):
        # Give the Workbench Version
        version = importlib.import_module("workbench").__version__
        cprint("lightpurple", f"Workbench Version: {version}")

        # Check the Workbench config
        self.cm = ConfigManager()
        if not self.cm.config_okay():
            # Invoke Onboarding Procedure
            onboard()

        # Our Metadata Object pull information from the Cloud Platform
        self.meta = None
        self.meta_status = "DIRECT"

        # Perform AWS connection test and other checks
        self.commands = dict()
        self.aws_status = self.check_aws_account()
        self.open_source_api_key = self.check_open_source_api_key()
        if self.aws_status:
            with silence_logs():
                self.import_workbench()

        # Try cached meta (if that fails it will be set to direct meta)
        self.try_cached_meta()

        # Register our custom commands
        self.commands["help"] = self.help
        self.commands["docs"] = self.doc_browser
        self.commands["summary"] = self.summary
        self.commands["incoming_data"] = self.incoming_data
        self.commands["glue_jobs"] = self.glue_jobs
        self.commands["data_sources"] = self.data_sources
        self.commands["feature_sets"] = self.feature_sets
        self.commands["models"] = self.models
        self.commands["endpoints"] = self.endpoints
        self.commands["pipelines"] = self.pipelines
        self.commands["log_debug"] = self.log_debug
        self.commands["log_trace"] = self.log_trace
        self.commands["log_info"] = self.log_info
        self.commands["log_important"] = self.log_important
        self.commands["log_warning"] = self.log_warning
        self.commands["config"] = self.show_config
        self.commands["status"] = self.status_description
        self.commands["plot"] = self.plot_manager
        self.commands["log"] = logging.getLogger("workbench")
        self.commands["get_meta"] = self.get_meta
        self.commands["params"] = importlib.import_module("workbench.api.parameter_store").ParameterStore()
        self.commands["secrets"] = importlib.import_module(
            "workbench.core.cloud_platform.aws.aws_secrets_manager"
        ).AWSSecretsManager()
        self.commands["df_store"] = importlib.import_module("workbench.api.df_store").DFStore()
        self.commands["inf_store"] = importlib.import_module("workbench_bridges.api.inference_store").InferenceStore()
        self.commands["graph_store"] = importlib.import_module("workbench.api.graph_store").GraphStore()
        self.commands["version"] = lambda: print(version)
        self.commands["cached_meta"] = self.switch_to_cached_meta
        self.commands["direct_meta"] = self.switch_to_direct_meta
        self.commands["log_theme"] = log_theme
        self.commands["reconnect"] = self.check_aws_account

        # Add cheminformatics utils if available
        if HAVE_CHEM_UTILS:
            self.commands["show"] = chem_utils.show

    def start(self):
        """Start the Workbench IPython shell"""
        cprint("magenta", "\nWelcome to Workbench!")
        if self.aws_status is False:
            cprint("red", "AWS Account Connection Failed...Review/Fix the Workbench Config:")
            cprint("red", f"Path: {self.cm.site_config_path}")
            self.show_config()
        else:
            self.help()
            self.summary()

        # Load the default IPython configuration
        config = load_default_config()
        config.TerminalInteractiveShell.autocall = 2
        config.TerminalInteractiveShell.prompts_class = WorkbenchPrompt
        config.TerminalInteractiveShell.highlighting_style_overrides = prompt_styles
        config.TerminalInteractiveShell.banner1 = ""

        # Merge custom commands and globals into the namespace
        locs = self.commands.copy()  # Copy the custom commands
        locs.update(globals())  # Merge with global namespace

        # Start IPython with the config and commands in the namespace
        try:
            if LooseVersion(IPython.__version__) >= LooseVersion("9.0.0"):
                ipython_argv = ["--no-tip", "--theme", "linux"]
            else:
                ipython_argv = []
            start_ipython(ipython_argv, user_ns=locs, config=config)
        finally:
            spinner = self.spinner_start("Goodbye to AWS:")
            with silence_logs():
                self.meta.close()
            spinner.stop()
            cprint("lightgreen", "Goodbye from Workbench!\n")

    def check_open_source_api_key(self) -> bool:
        """Check the current Configuration Status

        Returns:
            bool: True if Open Source API Key, False otherwise
        """
        config = self.cm.get_all_config()
        return config["API_KEY_INFO"]["license_id"] == "Open Source"

    @staticmethod
    def check_aws_account() -> bool:
        """Check if the AWS Account is Set up Correctly

        Returns:
            bool: True if AWS Account is set up correctly, False otherwise
        """
        cprint("lightgreen", "Checking AWS Account Connection...")
        try:
            try:
                aws_clamp = importlib.import_module(
                    "workbench.core.cloud_platform.aws.aws_account_clamp"
                ).AWSAccountClamp()
                aws_clamp.check_aws_identity()
                cprint("lightgreen", "AWS Account Check AOK!")
            except RuntimeError:
                cprint("red", "AWS Account Check Failed: Check AWS_PROFILE and/or Renew SSO Token...")
                return False
        except botocore.exceptions.ProfileNotFound:
            cprint("red", "AWS Account Check Failed: Check AWS_PROFILE...")
            return False
        except botocore.exceptions.NoCredentialsError:
            cprint("red", "AWS Account Check Failed: Check AWS Credentials...")
            return False

        # Okay assume everything is good
        return True

    def show_config(self):
        """Show the current Workbench Config"""
        cprint("yellow", "\nWorkbench Config:")
        cprint("lightblue", f"Path: {self.cm.site_config_path}")
        config = self.cm.get_all_config()
        for key, value in config.items():
            cprint(["lightpurple", "\t" + key, "lightgreen", value])

    def import_workbench(self):
        # Import all the Workbench modules
        spinner = self.spinner_start("Spinning up Workbench:")
        try:
            # These are the classes we want to expose to the REPL
            self.commands["DataSource"] = importlib.import_module("workbench.api.data_source").DataSource
            self.commands["FeatureSet"] = importlib.import_module("workbench.api.feature_set").FeatureSet
            self.commands["Model"] = importlib.import_module("workbench.api.model").Model
            self.commands["CachedModel"] = importlib.import_module("workbench.cached.cached_model").CachedModel
            self.commands["ModelType"] = importlib.import_module("workbench.api.model").ModelType
            self.commands["Endpoint"] = importlib.import_module("workbench.api.endpoint").Endpoint
            self.commands["Monitor"] = importlib.import_module("workbench.api.monitor").Monitor
            self.commands["ParameterStore"] = importlib.import_module("workbench.api.parameter_store").ParameterStore
            self.commands["DFStore"] = importlib.import_module("workbench.api.df_store").DFStore
            self.commands["GraphStore"] = importlib.import_module("workbench.api.graph_store").GraphStore
            self.commands["PandasToFeatures"] = importlib.import_module(
                "workbench.core.transforms.pandas_transforms"
            ).PandasToFeatures
            self.commands["Meta"] = importlib.import_module("workbench.api").Meta
            self.commands["CachedMeta"] = importlib.import_module("workbench.cached.cached_meta").CachedMeta
            self.commands["View"] = importlib.import_module("workbench.core.views.view").View
            self.commands["DisplayView"] = importlib.import_module("workbench.core.views.display_view").DisplayView
            self.commands["TrainingView"] = importlib.import_module("workbench.core.views.training_view").TrainingView
            self.commands["ComputationView"] = importlib.import_module(
                "workbench.core.views.computation_view"
            ).ComputationView
            self.commands["InferenceView"] = importlib.import_module(
                "workbench.core.views.inference_view"
            ).InferenceView
            self.commands["PandasToView"] = importlib.import_module("workbench.core.views.pandas_to_view").PandasToView
            self.commands["Pipeline"] = importlib.import_module("workbench.api.pipeline").Pipeline

            # Algorithms
            self.commands["FSP"] = importlib.import_module(
                "workbench.algorithms.dataframe.feature_space_proximity"
            ).FeatureSpaceProximity

            # These are 'nice to have' imports
            self.commands["pd"] = importlib.import_module("pandas")
            self.commands["wr"] = importlib.import_module("awswrangler")
            self.commands["pprint"] = importlib.import_module("pprint").pprint
        finally:
            spinner.stop()

    def help(self, *args):
        """Custom help command for the Workbench REPL

        Args:
            *args: Arguments passed to the help command.
        """
        # If we have args forward to the built-in help function
        if args:
            help(*args)

        # Otherwise show the Workbench help message
        else:
            cprint("lightblue", self.help_txt())

    @staticmethod
    def help_txt():
        help_msg = """    Commands:
        - help: Show this help message
        - docs: Open browser to show Workbench Documentation
        - data_sources: List all the DataSources in AWS
        - feature_sets: List all the FeatureSets in AWS
        - models: List all the Models in AWS
        - endpoints: List all the Endpoints in AWS
        - config: Show the current Workbench Config
        - status: Show the current Workbench Status
        - log_(debug/info/important/warning): Set the Workbench log level
        - exit: Exit Workbench REPL"""
        return help_msg

    def spinner_start(self, text: str, color: str = "lightpurple") -> Spinner:
        # Import all the Workbench modules
        spinner = Spinner(color, text)
        spinner.start()  # Start the spinner
        return spinner

    @staticmethod
    def doc_browser():
        """Open a browser and start the Dash app and open a browser."""
        url = "https://supercowpowers.github.io/workbench/"
        webbrowser.open(url)

    def summary(self):
        """Show a summary of all the AWS Artifacts"""

        # Grab information about all the AWS Artifacts
        spinner = self.spinner_start("Chatting with AWS:")
        try:
            # We're filling in Summary Data for all the AWS Services
            summary_data = {
                "INCOMING_DATA": self.meta.incoming_data(),
                "ETL_JOBS": self.meta.etl_jobs(),
                "DATA_SOURCES": self.meta.data_sources(),
                "FEATURE_SETS": self.meta.feature_sets(),
                "MODELS": self.meta.models(),
                "ENDPOINTS": self.meta.endpoints(),
            }
        finally:
            spinner.stop()

        # Print out the AWS Artifacts Summary
        cprint("yellow", "\nAWS Artifacts Summary:")
        for name, df in summary_data.items():
            # Pad the name to 15 characters
            name = (name + " " * 15)[:15]

            # Sanity check the dataframe
            if df.empty:
                examples = ""

            # Get the first three items in the first column
            else:
                examples = ", ".join(df.iloc[:, 0].tolist())
                if len(examples) > 70:
                    examples = examples[:70] + "..."

            # Print the summary
            cprint(["lightpurple", "\t" + name, "lightgreen", str(df.shape[0]) + "  ", "purple_blue", examples])

    def incoming_data(self):
        return self.meta.incoming_data()

    def glue_jobs(self):
        return self.meta.etl_jobs()

    def data_sources(self):
        return self.meta.data_sources()

    def feature_sets(self, details: bool = False):
        return self.meta.feature_sets(details=details)

    def models(self, details: bool = False):
        return self.meta.models(details=details)

    def endpoints(self):
        return self.meta.endpoints()

    def pipelines(self):
        return self.meta.pipelines()

    @staticmethod
    def log_debug():
        logging.getLogger("workbench").setLevel(logging.DEBUG)

    @staticmethod
    def log_trace():
        logging.getLogger("workbench").setLevel(TRACE_LEVEL_NUM)

    @staticmethod
    def log_info():
        logging.getLogger("workbench").setLevel(logging.INFO)

    @staticmethod
    def log_important():
        logging.getLogger("workbench").setLevel(IMPORTANT_LEVEL_NUM)

    @staticmethod
    def log_warning():
        logging.getLogger("workbench").setLevel(logging.WARNING)

    def status_lights(self) -> list[(Token, str)]:
        """Check the status of AWS, Redis, and API Key and return Token colors

        Returns:
            list[(Token, str)]: A list of Token colors and status symbols
        """
        _status_lights = [(Token.Blue, "[")]

        # AWS Account Status
        if self.aws_status:
            _status_lights.append((Token.Green, "●"))
        else:
            _status_lights.append((Token.Red, "●"))

        # Cached Meta Status
        if self.meta_status == "CACHED":
            _status_lights.append((Token.Green, "●"))
        elif self.meta_status == "DIRECT":
            _status_lights.append((Token.Blue, "●"))
        elif self.meta_status == "FAIL":
            _status_lights.append((Token.Orange, "●"))
        else:  # Unknown
            _status_lights.append((Token.Grey, "●"))

        # Check API Key
        if self.open_source_api_key:
            _status_lights.append((Token.Lightpurple, "●"))
        else:
            _status_lights.append((Token.Green, "●"))

        _status_lights.append((Token.Blue, "]"))

        return _status_lights

    def status_description(self):
        """Print a description of the status of AWS, Redis, and API Key"""

        # AWS Account
        if self.aws_status:
            cprint("lightgreen", "\t● AWS Account: OK")
        else:
            cprint("red", "\t● AWS Account: Failed to Connect")

        # Redis
        if self.meta_status == "CACHED":
            cprint("lightgreen", "\t● Meta: Cached")
        elif self.meta_status == "DIRECT":
            cprint("lightblue", "\t● Meta: Direct")
        elif self.meta_status == "FAIL":
            cprint("orange", "\t● Meta: Failed to Connect")

        # API Key
        if self.open_source_api_key:
            cprint("lightpurple", "\t● API Key: Open Source")
        else:
            cprint("lightgreen", "\t● API Key: Enterprise")

    # Helpers method to switch from direct Meta to Cached Meta
    def try_cached_meta(self):

        with silence_logs():
            self.meta = CachedMeta()
        if self.meta.check():
            self.meta_status = "CACHED"
            cprint("lightblue", "Using Cached Meta...")
        else:
            self.meta_status = "DIRECT"
            cprint("darkyellow", "Using Direct Meta [slower]...")
            with silence_logs():
                self.meta.close()
                self.meta = Meta()

    def switch_to_cached_meta(self):
        from workbench.api import Meta
        from workbench.cached.cached_meta import CachedMeta

        self.meta = CachedMeta()
        if self.meta.check():
            self.meta_status = "CACHED"
            cprint("lightblue", "Switched to Cached Meta...")
        else:
            self.meta.close()
            self.meta_status = "FAIL"
            cprint("orange", "Failed to Switch to Cached Meta...")
            cprint("darkyellow", "Using Direct Meta [slower]...")
            self.meta = Meta()

    def switch_to_direct_meta(self):
        from workbench.api import Meta

        # Close the current Meta object
        if self.meta:
            self.meta.close()
        # Create a new direct Meta object
        self.meta = Meta()
        self.meta_status = "DIRECT"
        cprint("darkyellow", "Switched to Direct Meta...")

    def get_meta(self):
        return self.meta

    def plot_manager(self, data, plot_type: str = "table", **kwargs):
        """Plot Manager for Workbench"""
        from workbench.web_interface.components.plugins import ag_table, graph_plot, scatter_plot

        if plot_type == "table":
            # Check what type of data we have
            if isinstance(data, pd.DataFrame):
                self.plot_plugin(ag_table.AGTable, data, **kwargs)
            else:
                # Does this object have a pull_dataframe() method?
                if hasattr(data, "pull_dataframe"):
                    data = data.pull_dataframe(limit=100)
                    self.plot_plugin(ag_table.AGTable, data, **kwargs)
                else:
                    cprint("yellow", f"Unknown Data Type for Table Plot '{type(data)}'")
        elif plot_type == "graph":
            self.plot_plugin(graph_plot.GraphPlot, data, **kwargs)
        elif plot_type == "scatter":
            self.plot_plugin(scatter_plot.ScatterPlot, data, **kwargs)
        else:
            cprint("yellow", f"Unknown Plot Type '{plot_type}'")

    @staticmethod
    def plot_plugin(plugin_class, data=None, **kwargs):
        """Plot data using a plugin.

        Args:
            plugin_class (PluginInterface): The plugin class to use.
            input_data (Optional): Optional input data (e.g., DataSource, FeatureSet, etc.)
            **kwargs: Additional keyword arguments for plugin properties.
                theme (str): The theme to use for the Dash app (default: "dark")
        """
        from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

        # Get kwargs
        theme = kwargs.get("theme", "dark")

        plugin_test = PluginUnitTest(plugin_class, theme=theme, input_data=data, **kwargs)

        # Run the server and open in the browser
        plugin_test.run()
        url = f"http://127.0.0.1:{plugin_test.port}"
        webbrowser.open(url)


# Launch Shell Entry Point
def launch_shell():
    global workbench_shell
    workbench_shell = WorkbenchShell()
    workbench_shell.start()


# Start the shell when running the script
if __name__ == "__main__":
    launch_shell()



"""Timing of pipeline endpoints vs. individual model endpoints"""

import re
import time

# Workbench imports
from workbench.api import FeatureSet, Endpoint
from workbench.utils.fast_inference import fast_inference

feature_set_name = "aqsol_features"
id_column = "id"

# Pull in a dataframe of smiles
fs = FeatureSet(feature_set_name)
df = fs.pull_dataframe()
total_df = df[[id_column, "smiles"]]

# Remove any rows with this SMILES substring "[I-].CN1C=CC=C\C1=C/[NH+]=O"
print(f"Before: {total_df.shape}")
# total_df = total_df[~total_df["smiles"].str.contains(r"[I-].CN1C=CC=C\C1=C/[NH+]=O")]
pattern = re.escape(r"[I-].CN1C=CC=C\C1=C/[NH+]=O")  # Escapes all special characters
total_df = total_df[~total_df["smiles"].str.contains(pattern)]
print(f"After: {total_df.shape}")

# First let's dump the information about all the endpoints
end_1 = Endpoint("tautomerize-v0-rt")
print(f"Endpoint: {end_1.name}, Instance: {end_1.instance_type}")
end_2 = Endpoint("smiles-to-taut-md-stereo-v0-rt")
print(f"Endpoint: {end_2.name}, Instance: {end_2.instance_type}")
end_3 = Endpoint("aqsol-mol-class-rt")
print(f"Endpoint: {end_3.name}, Instance: {end_3.instance_type}")
end_pipe = Endpoint("pipeline-model")
print(f"Endpoint: {end_pipe.name}, Instance: {end_pipe.instance_type}")
end_pipe_fast = Endpoint("pipeline-model-fast")
print(f"Endpoint: {end_pipe_fast.name}, Instance: {end_pipe_fast.instance_type}")

# We're going to grab our Sagemaker Session from an endpoint (the all give the same session)
session = end_1.sm_session

# We're going to run 3 timings on 10, 100, 500, 1000, 10000 rows
for n in [10, 100, 500, 1000, 10000]:
    input_df = total_df.head(n)
    print(f"Timing for {n} rows")

    # First time the individual endpoints

    # Tautomerize
    total_start = time.time()
    # end_1 = Endpoint("tautomerize-v0-rt")
    # df = end_1.fast_inference(input_df)
    df = fast_inference(end_1.name, input_df, session)
    time_taut = time.time() - total_start

    # Molecular Descriptors
    start = time.time()
    # end_2 = Endpoint("smiles-to-taut-md-stereo-v0-rt")
    # df = end_2.fast_inference(df)
    df = fast_inference(end_2.name, df, session)
    time_md = time.time() - start

    # AQSOL Classification
    start = time.time()
    # end_3 = Endpoint("aqsol-mol-class-rt")
    # df = end_3.fast_inference(df)
    df = fast_inference(end_3.name, df, session)
    time_model = time.time() - start
    total_time = time.time() - total_start

    print(f"Individual endpoints Total: {total_time} seconds")
    print(f"Taut: {time_taut} seconds, MD: {time_md} seconds, Model: {time_model} seconds")

    # Now time the pipeline endpoint
    start = time.time()
    # end_pipe = Endpoint("pipeline-model")
    # df = end_pipe.fast_inference(input_df)
    df = fast_inference(end_pipe.name, input_df, session)
    end = time.time()
    print(f"Pipeline endpoint: {end - start} seconds")

    # Now time the pipeline fast endpoint
    start = time.time()
    # end_pipe_fast = Endpoint("pipeline-model-fast")
    # df = end_pipe_fast.fast_inference(input_df)
    df = fast_inference(end_pipe_fast.name, input_df, session)
    end = time.time()
    print(f"Pipeline Fast endpoint: {end - start} seconds")



import sys
import time
import logging
from io import StringIO
import matplotlib.pyplot as plt

# SageMaker imports
from sagemaker.serializers import CSVSerializer
from sagemaker.deserializers import CSVDeserializer
from sagemaker import Predictor

# Workbench imports
from workbench.api import FeatureSet, Model, Endpoint

log = logging.getLogger("workbench")


def plot_timings(timings):
    """Plots the inference timings."""
    x = [1, 10, 100, 1000, 10000]
    plt.figure(figsize=(10, 6))

    plt.plot(x, timings["Workbench Serverless"], label="Workbench Serverless", color="blue")
    plt.plot(x, timings["Workbench Realtime"], label="Workbench Realtime", color="green")
    plt.plot(x, timings["AWS Serverless"], label="AWS Serverless", color="red")
    plt.plot(x, timings["AWS Realtime"], label="AWS Realtime", color="orange")

    plt.xscale("log")
    plt.xlabel("Number of Rows")
    plt.ylabel("Inference Time (seconds)")
    plt.title("Inference Timing Comparison")
    plt.legend()
    plt.grid(True)
    plt.show()


if __name__ == "__main__":

    plot = True  # Set this flag to control plotting

    # Check that the two Endpoints exist
    if not Endpoint("test-timing-serverless").exists():
        log.error("The test-timing-serverless endpoint does not exist. Please create it first.")
        sys.exit(1)

    if not Endpoint("test-timing-realtime").exists():
        log.error("The test-timing-realtime endpoint does not exist. Please create it first.")
        sys.exit(1)

    # Timing results storage
    timings = {
        "Workbench Serverless": [],
        "Workbench Realtime": [],
        "AWS Serverless": [],
        "AWS Realtime": [],
    }

    # Get the endpoints
    start_time = time.time()
    serverless = Endpoint("test-timing-serverless")
    print(f"Workbench Serverless Endpoint: Construction Time: {time.time() - start_time}")

    start_time = time.time()
    realtime = Endpoint("test-timing-realtime")
    print(f"Workbench Realtime Endpoint: Construction Time: {time.time() - start_time}")

    # Backtrace the endpoints to get the FeatureSet
    fs = FeatureSet(Model(serverless.get_input()).get_input())

    # Grab the FeatureSet data
    data = fs.pull_dataframe()

    #
    # Done with the data pull (we're not including data pull as part of the timing)
    #

    # Now we're going to skip any Workbench code and just use the AWS SDK
    print("\n*** Timing Inference using JUST the AWS SDK ***")

    # Predictor creation
    start_time = time.time()
    serverless_predictor = Predictor(
        "test-timing-serverless",
        sagemaker_session=serverless.sm_session,
        serializer=CSVSerializer(),
        deserializer=CSVDeserializer(),
    )
    realtime_predictor = Predictor(
        "test-timing-realtime",
        sagemaker_session=realtime.sm_session,
        serializer=CSVSerializer(),
        deserializer=CSVDeserializer(),
    )
    print(f"Sagemaker Predictor: Construction Time: {time.time() - start_time}")

    # Just AWS: Collect inference times on 1, 10, 100, 1000, and 10000 rows
    for i in [1, 10, 100, 1000, 10000]:
        data_sample = data.sample(n=i, replace=True)
        print(f"\nTiming Inference on {len(data_sample)} rows")

        # Convert into a CSV buffer and send it to the AWS serverless predictor
        start_time = time.time()
        csv_buffer = StringIO()
        data_sample.to_csv(csv_buffer, index=False)
        serverless_predictor.predict(csv_buffer.getvalue())
        aws_serverless_time = time.time() - start_time
        timings["AWS Serverless"].append(aws_serverless_time)
        print(f"\tServerless Inference Time: {aws_serverless_time}")

        # Convert into a CSV buffer and send it to the AWS realtime predictor
        start_time = time.time()
        csv_buffer = StringIO()
        data_sample.to_csv(csv_buffer, index=False)
        realtime_predictor.predict(csv_buffer.getvalue())
        aws_realtime_time = time.time() - start_time
        timings["AWS Realtime"].append(aws_realtime_time)
        print(f"\tRealtime Inference Time: {aws_realtime_time}")

    print("\n*** Timing Inference using Workbench API ***")

    # Collect inference times on 1, 10, 100, 1000, and 10000 rows
    for i in [1, 10, 100, 1000, 10000]:
        data_sample = data.sample(n=i, replace=True)
        print(f"\nTiming Inference on {len(data_sample)} rows")
        start_time = time.time()
        serverless.fast_inference(data_sample)
        serverless_time = time.time() - start_time
        timings["Workbench Serverless"].append(serverless_time)
        print(f"\tServerless Inference Time: {serverless_time}")

        start_time = time.time()
        realtime.fast_inference(data_sample)
        realtime_time = time.time() - start_time
        timings["Workbench Realtime"].append(realtime_time)
        print(f"\tRealtime Inference Time: {realtime_time}")

    # Plot timings if the flag is set
    if plot:
        plot_timings(timings)



import awswrangler as wr
from workbench.api import Meta


def delete_invalid_views(database_scope):
    """Delete invalid views from the AWS Data Catalog"""

    # Create an instance of the Meta class
    meta = Meta()

    # Grab the boto3 session from the DataCatalog
    boto3_session = meta.boto3_session

    # Okay now delete invalid views for each database
    for database in database_scope:
        # Get all the views in this database
        views = meta.views(database)

        # Query the view, if we get an 'INVALID_VIEW' error, delete the view
        for view in views["Name"]:
            try:
                # Attempt a simple query to check if the view is valid
                print(f"Checking view: {view}...")
                query = f'SELECT * FROM "{view}" limit 0'
                wr.athena.read_sql_query(sql=query, database=database, ctas_approach=False, boto3_session=boto3_session)
            except Exception as e:
                if "INVALID_VIEW" in str(e):
                    print(f"\tDELETING invalid view: {view}...")
                    wr.catalog.delete_table_if_exists(database, view, boto3_session=boto3_session)
                else:
                    print(f"Error querying view {view}: {e}")


if __name__ == "__main__":
    database_scope = ["workbench", "sagemaker_featurestore"]
    delete_invalid_views(database_scope)



"""This Script creates a Realtime Endpoint for testing

Endpoints:
    - abalone-regression-end-rt
"""

import logging

# Workbench Imports
from workbench.api.model import Model
from workbench.api.endpoint import Endpoint

# Setup the logger
log = logging.getLogger("workbench")

if __name__ == "__main__":

    # Recreate Flag in case you want to recreate the artifacts
    recreate = False

    # Create the abalone_regression Endpoint
    if recreate or not Endpoint("abalone-regression-rt").exists():
        m = Model("abalone-regression")
        m.to_endpoint("abalone-regression-rt", tags=["abalone", "regression", "realtime"], serverless=False)



import os
import logging
from datetime import datetime, timezone
import getpass
import time  # For managing send intervals

# Workbench imports
from workbench_bridges.utils.execution_environment import (
    running_on_lambda,
    running_on_glue,
    running_on_ecs,
    running_on_docker,
    glue_job_name,
    ecs_job_name,
)
from workbench.core.cloud_platform.aws.aws_session import AWSSession


class CloudWatchHandler(logging.Handler):
    """A helper class to add a CloudWatch Logs handler with buffering to a logger"""

    def __init__(self, buffer_size=10, send_interval=5):
        super().__init__()  # Initialize the base Handler class

        # Buffer to hold log messages
        self.buffer = []
        self.buffer_size = buffer_size
        self.send_interval = send_interval
        self.last_sent_time = time.time()

        # Get our session and log stream name
        self.boto3_session = AWSSession().boto3_session
        self.log_stream_name = self.determine_log_stream()
        self.cloudwatch_client = self.boto3_session.client("logs")
        self.sequence_token = None
        self.log_group_name = "WorkbenchLogGroup"

        # Create the log stream
        self.create_log_stream()

    def emit(self, record):
        """Add a log message to the buffer and send when ready"""
        message = self.format(record)
        log_event = {"timestamp": int(record.created * 1000), "message": message}
        self.buffer.append(log_event)

        # Check if the buffer is full or if the time interval has passed
        if len(self.buffer) >= self.buffer_size or (time.time() - self.last_sent_time) >= self.send_interval:
            self.send_logs()

    def send_logs(self):
        """Send buffered log messages to CloudWatch"""
        if not self.buffer:
            return  # Nothing to send

        # Filter out logs older than 24 hours
        time_delta = 24 * 60 * 60 * 1000  # 24 hours in milliseconds
        now = int(time.time() * 1000)  # Current time in milliseconds
        self.buffer = [event for event in self.buffer if now - event["timestamp"] <= time_delta]

        if not self.buffer:
            return  # After filtering, no valid logs to send

        # Sort the buffer by timestamp to ensure chronological order
        self.buffer.sort(key=lambda event: event["timestamp"])

        log_event = {
            "logGroupName": self.log_group_name,
            "logStreamName": self.log_stream_name,
            "logEvents": self.buffer,
        }

        if self.sequence_token:
            log_event["sequenceToken"] = self.sequence_token

        try:
            response = self.cloudwatch_client.put_log_events(**log_event)
            self.sequence_token = response.get("nextSequenceToken")
            self.buffer.clear()  # Clear the buffer after successful send
        except self.cloudwatch_client.exceptions.InvalidSequenceTokenException as e:
            # Handle sequence token exceptions by updating the token and retrying
            self.sequence_token = e.response["Error"]["Message"].split()[-1]
            log_event["sequenceToken"] = self.sequence_token
            self.cloudwatch_client.put_log_events(**log_event)  # Retry without recursion
            self.buffer.clear()
        except Exception as e:
            logging.error(f"Failed to send logs to CloudWatch: {e}")

        # Update last sent time after successful send
        self.last_sent_time = time.time()

    def flush(self):
        """Ensure all logs are sent"""
        self.send_logs()  # Flush remaining logs in the buffer

    def create_log_stream(self):
        """Create CloudWatch Log Stream if it doesn't exist"""
        try:
            self.cloudwatch_client.create_log_stream(
                logGroupName=self.log_group_name, logStreamName=self.log_stream_name
            )
        except self.cloudwatch_client.exceptions.ResourceAlreadyExistsException:
            pass

    def determine_log_stream(self):
        """Determine the log stream name based on the environment."""
        if running_on_lambda():
            job_name = os.environ.get("AWS_LAMBDA_FUNCTION_NAME", "unknown")
            return f"lambda/{job_name}"
        elif running_on_glue():
            job_name = glue_job_name()
            unique_id = self.get_unique_identifier(job_name)
            return f"glue/{job_name}/{unique_id}"
        elif running_on_ecs():
            job_name = ecs_job_name()
            return f"ecs/{job_name}"
        elif running_on_docker():
            return "docker"
        else:
            return f"laptop/{getpass.getuser()}"

    def get_unique_identifier(self, job_name):
        """Get a unique identifier for the log stream."""
        return datetime.now(timezone.utc).strftime("%Y_%m_%d_%H_%M_%S")


if __name__ == "__main__":
    # Example usage
    logger = logging.getLogger("Workbench")
    logger.setLevel(logging.INFO)
    cloudwatch_handler = CloudWatchHandler()
    logger.addHandler(cloudwatch_handler)
    logger.info("Test log message to CloudWatch")

    # Test the get_unique_identifier function
    print(cloudwatch_handler.get_unique_identifier("Glue_Job_1"))



import argparse
import logging
from workbench.utils.redis_cache import RedisCache

# Set up logging
log = logging.getLogger("workbench")


def main(days, dry_run):
    # Create an instance of RedisCache
    redis_cache = RedisCache()

    # Check if Redis is available
    if not redis_cache.check():
        log.error("Redis not available, exiting...")
        return

    # Call the delete_keys_older_than method with the specified parameters
    redis_cache.delete_keys_older_than(days=days, dry_run=dry_run)


if __name__ == "__main__":
    # Argument parsing
    parser = argparse.ArgumentParser(description="Delete keys older than N days from Redis.")
    parser.add_argument(
        "--days",
        type=int,
        default=30,
        help="Number of days after which keys are deleted (default: 30)",
    )
    parser.add_argument(
        "--no-dry-run",
        action="store_false",
        dest="dry_run",
        help="Actually delete keys (the default is to show what would be deleted)",
    )

    args = parser.parse_args()

    # Execute main function
    main(days=args.days, dry_run=args.dry_run)



"""FeaturesToPandas: Class to transform a FeatureSet into a Pandas DataFrame"""

import pandas as pd

# Local imports
from workbench.core.transforms.transform import Transform, TransformInput, TransformOutput
from workbench.core.artifacts.feature_set_core import FeatureSetCore


class FeaturesToPandas(Transform):
    """FeaturesToPandas: Class to transform a FeatureSet into a Pandas DataFrame

    Common Usage:
        ```python
        feature_to_df = FeaturesToPandas(feature_set_name)
        feature_to_df.transform(max_rows=<optional max rows to sample>)
        my_df = feature_to_df.get_output()
        ```
    """

    def __init__(self, feature_set_name: str):
        """FeaturesToPandas Initialization"""

        # Call superclass init
        super().__init__(input_name=feature_set_name, output_name="DataFrame")

        # Set up all my instance attributes
        self.input_type = TransformInput.FEATURE_SET
        self.output_type = TransformOutput.PANDAS_DF
        self.output_df = None
        self.transform_run = False

    def transform_impl(self, max_rows=100000):
        """Convert the FeatureSet into a Pandas DataFrame"""

        # Grab the Input (Feature Set)
        input_data = FeatureSetCore(self.input_name)
        if not input_data.exists():
            self.log.critical(f"Feature Set Check on {self.input_name} failed!")
            return

        # Grab the table for this Feature Set
        table = input_data.athena_table

        # Get the list of columns (and subtract metadata columns that might get added)
        columns = input_data.columns
        filter_columns = ["write_time", "api_invocation_time", "is_deleted"]
        columns = ", ".join([x for x in columns if x not in filter_columns])

        # Get the number of rows in the Feature Set
        num_rows = input_data.num_rows()

        # If the data source has more rows than max_rows, do a sample query
        if num_rows > max_rows:
            percentage = round(max_rows * 100.0 / num_rows)
            self.log.important(f"DataSource has {num_rows} rows.. sampling down to {max_rows}...")
            query = f'SELECT {columns} FROM "{table}" TABLESAMPLE BERNOULLI({percentage})'
        else:
            query = f'SELECT {columns} FROM "{table}"'

        # Mark the transform as complete and set the output DataFrame
        self.transform_run = True
        self.output_df = input_data.query(query)

    def post_transform(self, **kwargs):
        """Post-Transform: Any checks on the Pandas DataFrame that need to be done"""
        self.log.info("Post-Transform: Checking Pandas DataFrame...")
        self.log.info(f"DataFrame Shape: {self.output_df.shape}")

    def get_output(self) -> pd.DataFrame:
        """Get the DataFrame Output from this Transform"""
        if not self.transform_run:
            self.transform()
        return self.output_df


if __name__ == "__main__":
    """Exercise the FeaturesToPandas Class"""

    # Setup Pandas output options
    pd.set_option("display.max_colwidth", 15)
    pd.set_option("display.max_columns", 15)
    pd.set_option("display.width", 1000)

    # Create the FeatureSet to DF Transform
    feature_to_df = FeaturesToPandas("test_features")

    # Transform the DataSource into a Pandas DataFrame (with max_rows = 1000)
    feature_to_df.transform(max_rows=1000)

    # Grab the output and show it
    my_df = feature_to_df.get_output()
    print(my_df)



"""iPython Utilities"""

from IPython import get_ipython
from IPython.display import display


def is_running_in_ipython() -> bool:
    """Check if the code is running inside a Jupyter notebook or iPython shell"""
    try:
        return get_ipython() is not None
    except NameError:
        return False


def display_error_and_raise(error_message: str) -> None:
    """Display an error message and raise an exception"""
    display(error_message)
    raise Exception(error_message)


if __name__ == "__main__":
    """Exercise the iPython Utilities"""
    if is_running_in_ipython():
        display_error_and_raise("This is an error message!")
    else:
        print("Not running in iPython!")



import awswrangler as wr
from workbench.api.data_source import DataSource
from pprint import pprint


def column_changes_test():
    # Create a new Data Source from an S3 Path
    source_path = "s3://workbench-public-data/common/abalone.csv"
    df = wr.s3.read_csv(source_path)
    df.rename(columns={"Diameter": "old"}, inplace=True)
    my_data = DataSource(df, "test_columns")
    pprint(my_data.summary())

    # Change the columns of the Data Source and create DataSource with same name
    df.rename(columns={"old": "new"}, inplace=True)
    my_data = DataSource(df, "test_columns")
    pprint(my_data.summary())


if __name__ == "__main__":
    column_changes_test()



"""Workbench Dashboard: A Workbench Web Application for viewing and managing Workbench Artifacts"""

from dash import Dash, html, dcc, page_container
import dash_bootstrap_components as dbc
from dash import page_registry

# Workbench Imports
from workbench.utils.plugin_manager import PluginManager
from workbench.utils.theme_manager import ThemeManager

# Set up the logging
import logging

log = logging.getLogger("workbench")


# Note: The 'app' and 'server' objects need to be at the top level since NGINX/uWSGI needs to
#       import this file and use the server object as an ^entry-point^ into the Dash Application Code

# Set up the Theme Manager
tm = ThemeManager()
tm.set_theme("auto")
css_files = tm.css_files()
print(css_files)

# Set the Dash App Title
app_title = tm.branding().get("app_title", "Workbench Dashboard")

# Create the Dash app
app = Dash(
    __name__,
    title=app_title,
    use_pages=True,
    external_stylesheets=css_files,
)

# Register the CSS route in the ThemeManager
tm.register_css_route(app)

# Note: The 'server' object is required for running the app with NGINX/uWSGI
server = app.server

# For Multi-Page Applications, we need to create a 'page container' to hold all the pages
plugin_info_id = "plugin-pages-info"
app.layout = html.Div(
    [
        # URL for subpage navigation (jumping to feature_sets, models, etc.)
        dcc.Location(id="url", refresh="callback-nav"),
        dcc.Store(id=plugin_info_id, data={}),
        dbc.Container([page_container], fluid=True, className="dbc dbc-ag-grid"),
    ],
    **{"data-bs-theme": tm.data_bs_theme()},
)

# Spin up the Plugin Manager
pm = PluginManager()

# Grab any plugin pages
plugin_pages = pm.get_pages()

# Setup each if the plugin pages (call layout and callbacks internally)
for name, page in plugin_pages.items():
    # Note: We need to catch any exceptions here
    try:
        page.page_setup(app)
    except Exception as e:
        # Put the exception full stack trace
        log.critical(f"Error setting up plugin page: {name}")
        log.critical(e, exc_info=True)
        continue
log.info("Done with Plugin Pages")

# Grab our plugin page info from the page registry and populate our plugin-pages-info store
dashboard_page_paths = [
    "/",
    "/data_sources",
    "/feature_sets",
    "/models",
    "/endpoints",
    "/pipelines",
    "/license",
    "/status",
]

# Pull the plugin pages path and name
plugin_pages = {}
for page_id, page_info in page_registry.items():
    if page_info["path"] not in dashboard_page_paths:
        plugin_pages[page_info["path"]] = page_info["name"]
print(plugin_pages)

# Update the plugin-pages-info store
for component in app.layout.children:
    if isinstance(component, dcc.Store) and component.id == plugin_info_id:
        component.data = plugin_pages


if __name__ == "__main__":
    """Run our web application in TEST mode"""
    # Note: This 'main' is purely for running/testing locally
    # app.run(host="0.0.0.0", port=8000, debug=True)
    app.run(host="0.0.0.0", port=8000)



"""This Script creates the Workbench Artifacts in AWS needed for the tests

DataSources:
    - test_data
    - abalone_data
FeatureSets:
    - test_features
    - abalone_features
Models:
    - abalone-regression
Endpoints:
    - abalone-regression-end
"""

import sys
from pathlib import Path
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
from workbench.api.data_source import DataSource
from workbench.api.feature_set import FeatureSet
from workbench.api.model import Model, ModelType
from workbench.api.endpoint import Endpoint


def redis_check():
    """Check if the Redis Database is available"""
    print("*** Redis Database Check ***")
    try:
        from workbench.utils.redis_cache import RedisCache

        RedisCache(prefix="test")
        print("Redis Database Check Success...")
    except RuntimeError as err:
        print(f"Redis Database Check Failed: {err} but this is fine.. Redis is optional")


if __name__ == "__main__":
    # Get the path to the dataset in the repository data directory
    test_data_path = Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "test_data.csv"
    abalone_data_path = Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "abalone.csv"

    """Check if the AWS Account is Setup Correctly"""
    print("*** AWS Identity Check ***")
    AWSAccountClamp().check_aws_identity()

    # Check that the Redis Database is available
    redis_check()

    # Create the abalone_data DataSource
    if not DataSource("abalone_data").exists():
        DataSource(abalone_data_path, name="abalone_data")

    # Create the abalone_features FeatureSet
    if not FeatureSet("abalone_features").exists():
        ds = DataSource("abalone_data")
        ds.to_features("abalone_features")

    # Create the abalone_regression Model
    if not Model("abalone-regression").exists():
        fs = FeatureSet("abalone_features")
        fs.to_model(
            name="abalone-regression",
            model_type=ModelType.REGRESSOR,
            target_column="class_number_of_rings",
            tags=["abalone", "regression"],
            description="Abalone Regression Model",
        )

    # Create the abalone_regression Endpoint
    if not Endpoint("abalone-regression").exists():
        model = Model("abalone-regression")
        model.to_endpoint(name="abalone-regression", tags=["abalone", "regression"])



import signal
import atexit
import sys
import time
import argparse
from datetime import datetime, timedelta, timezone

# Workbench Imports
from workbench.utils.repl_utils import cprint, Spinner
from workbench.utils.cloudwatch_utils import get_cloudwatch_client, get_active_log_streams, stream_log_events

# Define the log levels to include all log levels above the specified level
log_level_map = {
    "all": [],
    "important": ["important", "warning", "monitor", "error", "critical"],
    "warning": ["warning", "monitor", "error", "critical"],
    "monitor": ["monitor", "error", "critical"],
    "error": ["error", "critical"],
}

# Global flag to display timestamps in local time
local_time = True


def date_display(dt):
    """Convert datetime to a concise human-readable format

    Args:
        dt (datetime): The datetime object to format.
    """
    if local_time:
        dt = dt.astimezone()
        return dt.strftime("%Y-%m-%d %I:%M%p")
    else:
        return dt.strftime("%Y-%m-%d %I:%M%p") + "(UTC)"


def get_latest_log_events(client, log_group_name, start_time, end_time=None, stream_filter=None):
    """Retrieve the latest log events from the active/filtered log streams in a CloudWatch Logs group."""

    # Initialize first run attribute
    if not hasattr(get_latest_log_events, "first_run"):
        get_latest_log_events.first_run = True

    log_events = []
    start_time_ms = int(start_time.timestamp() * 1000)

    # Use the util function to get active streams
    active_streams = get_active_log_streams(log_group_name, start_time_ms, stream_filter, client)

    if active_streams:
        print(f"Active log streams: {len(active_streams)}")
        for stream in active_streams:
            print(f"\t - {stream}")
        print(f"Processing log events from {date_display(start_time)} on {len(active_streams)} active log streams...")
        get_latest_log_events.first_run = False
    else:
        if get_latest_log_events.first_run:
            print(f"No active log streams (start_time:{date_display(start_time)}, stream-filter:'{stream_filter}')")
            get_latest_log_events.first_run = False
        else:
            print("Monitoring for new events...")
        return log_events

    # Use the util function to stream events from each log stream
    for log_stream_name in active_streams:
        spinner = Spinner("lightpurple", f"Pulling events from {log_stream_name}:")
        spinner.start()
        log_stream_events = 0

        # Stream events using the util function
        for event in stream_log_events(
            log_group_name, log_stream_name, start_time, end_time, follow=False, client=client
        ):
            log_stream_events += 1
            log_events.append(event)

        spinner.stop()
        print(f"Processed {log_stream_events} events from {log_stream_name} (Total: {len(log_events)})")

    return log_events


def merge_ranges(ranges):
    """Merge overlapping or adjacent line ranges."""
    if not ranges:
        return []

    # Sort ranges by start line
    ranges.sort(key=lambda x: x[0])
    merged = [ranges[0]]

    for current in ranges[1:]:
        last = merged[-1]
        if current[0] <= last[1] + 1:
            # If the current range overlaps or is adjacent to the last range, merge them
            merged[-1] = (last[0], max(last[1], current[1]))
        else:
            merged.append(current)

    return merged


def monitor_log_group(
    log_group_name,
    start_time,
    end_time=None,
    poll_interval=10,
    log_level=None,
    search_terms=None,
    before=10,
    after=0,
    stream_filter=None,
):
    """Continuously monitor the CloudWatch Logs group for new log messages from all log streams."""
    client = get_cloudwatch_client()

    # Convert any search terms to lowercase
    search_terms = [term.lower() for term in search_terms] if search_terms else []

    # Handle log levels (typically if a search term is used we should use 'all' log levels
    if search_terms:
        log_level = "all"
    log_levels = log_level_map.get(log_level.lower(), [log_level]) if log_level else []

    print(f"Monitoring log group: {log_group_name} from {date_display(start_time)}")
    print(f"Log levels: {log_levels}")
    print(f"Search terms: {search_terms}")

    while True:
        # Get the latest log events with stream filtering if provided
        all_log_events = get_latest_log_events(client, log_group_name, start_time, end_time, stream_filter)

        # Match log levels and search terms, collect ranges, and merge overlapping ranges
        ranges = []
        for i, event in enumerate(all_log_events):
            # Match any of log levels in the log message
            if not log_levels or any(term in event["message"].lower() for term in log_levels):

                # Check the search terms
                if not search_terms or any(term in event["message"].lower() for term in search_terms):
                    # Calculate the start and end index for this match
                    start_index = max(i - before, 0)
                    end_index = min(i + after, len(all_log_events) - 1)
                    ranges.append((start_index, end_index))

        # Merge overlapping ranges
        merged_ranges = merge_ranges(ranges)

        # Collect filtered events based on merged ranges
        filtered_events = []
        for start, end in merged_ranges:
            filtered_events.extend(all_log_events[start : end + 1])

            # These are just blank lines to separate the log message 'groups'
            filtered_events.append({"logStreamName": None, "timestamp": None, "message": ""})
            filtered_events.append({"logStreamName": None, "timestamp": None, "message": ""})

        # Display the filtered log events
        if not filtered_events:
            print("No log events found, matching the specified criteria...")
        for event in filtered_events:
            if event["logStreamName"] is None and event["timestamp"] is None:
                print("")
            else:
                log_stream_name = event["logStreamName"]
                timestamp = datetime.fromtimestamp(event["timestamp"] / 1000, tz=timezone.utc)
                message = event["message"].strip()
                print(f"[{log_stream_name}] [{date_display(timestamp)}] {message}")

        # Update the start time to just after the last event's timestamp
        if end_time is None:
            start_time = datetime.now(timezone.utc)
        else:
            break

        # Wait for the next poll if monitoring realtime logs
        if end_time is None:
            time.sleep(poll_interval)
        else:
            break


# Function to handle SIGINT (Control-C)
def signal_handler(sig, frame):
    sys.exit(0)


# Function to handle cleanup tasks at exit
def cleanup():
    cprint("lightgreen", "\nGoodbye from Workbench CloudWatch!")


def parse_args():
    parser = argparse.ArgumentParser(description="Monitor CloudWatch Logs.")
    parser.add_argument(
        "--start-time",
        type=int,
        default=600,
        help="Start time in minutes ago. Default is 600 minutes ago.",
    )
    parser.add_argument(
        "--end-time",
        type=int,
        help="End time in minutes ago for fetching a range of logs.",
    )
    parser.add_argument(
        "--poll-interval",
        type=int,
        default=60,
        help="Polling interval in seconds. Default is 60 seconds.",
    )
    parser.add_argument(
        "--utc-time",
        action="store_true",
        help="Display timestamps in UTC instead of local.",
    )
    parser.add_argument("--log-level", default="error", help="Log level to filter log messages.")
    parser.add_argument("--search", nargs="+", help="Search terms to filter log messages.")
    parser.add_argument(
        "--before",
        type=int,
        default=10,
        help="Number of lines to include before the log level match.",
    )
    parser.add_argument(
        "--after",
        type=int,
        default=0,
        help="Number of lines to include after the log level match.",
    )
    parser.add_argument("--stream", help="Filter log streams by a substring.")

    return parser.parse_args()


def main():
    args = parse_args()

    # Register the signal handler for SIGINT (Control-C)
    signal.signal(signal.SIGINT, signal_handler)

    # Register the cleanup function to be called at exit
    atexit.register(cleanup)

    # Determine the start time and end time for log monitoring (in UTC)
    start_time = datetime.now(timezone.utc) - timedelta(minutes=args.start_time)
    end_time = datetime.now(timezone.utc) - timedelta(minutes=args.end_time) if args.end_time else None

    # Set the global flag to display timestamps in local time
    global local_time
    local_time = not args.utc_time

    # Monitor the CloudWatch Logs group
    monitor_log_group(
        "WorkbenchLogGroup",
        start_time=start_time,
        end_time=end_time,
        poll_interval=args.poll_interval,
        log_level=args.log_level,
        search_terms=args.search,
        before=args.before,
        after=args.after,
        stream_filter=args.stream,
    )


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, auc

# Generate more accurate fake data with correlation between confidence and accuracy
np.random.seed(42)
n_samples = 1000
actual_values = np.random.uniform(1.5, 4.5, n_samples)

# Split actual_values into three parts
split_index1 = int(n_samples / 3)
split_index2 = int(2 * n_samples / 3)

actual_values_high = actual_values[:split_index1]
actual_values_medium = actual_values[split_index1:split_index2]
actual_values_low = actual_values[split_index2:]

# High confidence -> low error
high_confidence_predictions = actual_values_high + np.random.normal(0, 0.1, split_index1)
high_confidence_scores = np.random.uniform(0.8, 1.0, split_index1)

# Medium confidence -> moderate error
medium_confidence_predictions = actual_values_medium + np.random.normal(0, 0.3, split_index2 - split_index1)
medium_confidence_scores = np.random.uniform(0.5, 0.8, split_index2 - split_index1)

# Low confidence -> high error
low_confidence_predictions = actual_values_low + np.random.normal(0, 0.5, n_samples - split_index2)
low_confidence_scores = np.random.uniform(0.0, 0.5, n_samples - split_index2)

# Combine all predictions and confidence scores
predictions = np.concatenate(
    [
        high_confidence_predictions,
        medium_confidence_predictions,
        low_confidence_predictions,
    ]
)
confidence_scores = np.concatenate([high_confidence_scores, medium_confidence_scores, low_confidence_scores])

# Define accuracy threshold (e.g., absolute error < 0.5)
accuracy_threshold = 0.5
absolute_errors = np.abs(predictions - np.concatenate([actual_values_high, actual_values_medium, actual_values_low]))
accurate = absolute_errors < accuracy_threshold

# Categorize confidence scores into high, medium, and low
confidence_bins = np.percentile(confidence_scores, [33, 66])
low_confidence = confidence_scores <= confidence_bins[0]
medium_confidence = (confidence_scores > confidence_bins[0]) & (confidence_scores <= confidence_bins[1])
high_confidence = confidence_scores > confidence_bins[1]


# Compute ROC curve and AUC for each category
def compute_roc(accurate, confidence_category):
    fpr, tpr, _ = roc_curve(accurate, confidence_category)
    roc_auc = auc(fpr, tpr)
    return fpr, tpr, roc_auc


fpr_low, tpr_low, roc_auc_low = compute_roc(accurate, low_confidence)
fpr_medium, tpr_medium, roc_auc_medium = compute_roc(accurate, medium_confidence)
fpr_high, tpr_high, roc_auc_high = compute_roc(accurate, high_confidence)

# Plot ROC curves
plt.figure()
plt.plot(
    fpr_low,
    tpr_low,
    color="red",
    lw=2,
    label=f"Low Confidence ROC (area = {roc_auc_low:.2f})",
)
plt.plot(
    fpr_medium,
    tpr_medium,
    color="orange",
    lw=2,
    label=f"Medium Confidence ROC (area = {roc_auc_medium:.2f})",
)
plt.plot(
    fpr_high,
    tpr_high,
    color="green",
    lw=2,
    label=f"High Confidence ROC (area = {roc_auc_high:.2f})",
)
plt.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--")
plt.xlim([0.0, 1.0])
plt.ylim([0.0, 1.05])
plt.xlabel("False Positive Rate")
plt.ylabel("True Positive Rate")
plt.title("ROC Curves for Different Confidence Levels")
plt.legend(loc="lower right")
plt.show()



"""Status:  A Workbench Web Interface to view Status Details about the Workbench Dashboard"""

from dash import register_page
from dash import html

# Workbench Imports
from workbench.web_interface.components.plugins import dashboard_status
from workbench.utils.config_manager import ConfigManager


# Register this page with Dash
register_page(
    __name__,
    path="/status",
    name="Workbench - Dashboard Status",
)

# Grab the Workbench ConfigManager
cm = ConfigManager()
config_details = cm.get_all_config()

# Create a Markdown component to display the license details
markdown_details = dashboard_status.DashboardStatus()
details_component = markdown_details.create_component("status_details")
updated_properties = markdown_details.update_properties(config_details)

# Set the properties directly on the server side before initializing the layout
for (component_id, prop), value in zip(markdown_details.properties, updated_properties):
    for child in details_component.children:
        if child.id == component_id:
            setattr(child, prop, value)
            break

# Simple layout for the license details
layout = html.Div(
    children=[
        html.H2("Workbench Dashboard Status"),
        details_component,
    ],
    style={"padding": "12px 30px 30px 30px"},
)



"""Unicode Symbols for Workbench"""

# A Dictionary/Map of Health Tags to Symbols
health_icons = {
    "failed": "🔴",
    "5xx_errors": "🔴",
    "no_model": "🔴",
    "5xx_errors_min": "🟠",
    "no_endpoint": "🟡",
    "model_type_unknown": "⚪",
    "model_not_found": "⚫",
    "metrics_needed": "🟣",
    "needs_onboard": "🔵",
    "healthy": "🟢",
    "unknown_error": "⚫",
    "no_activity": "➖",
}


def tag_symbols(tag_list: str) -> str:
    """Return the symbols for the given list of tags"
    Args:
        tag_list (str): A string of tags separated by :
    Returns:
        str: The symbol for the given tag (or "" if no symbol)
    """

    # Split the tag list and return the symbol
    symbol_list = []
    tag_list = tag_list.split(":")

    # Special case for empty tag list
    if len(tag_list) == 0 or tag_list[0] == "":
        return health_icons["healthy"]

    # Special case for no_activity with no other tags
    if len(tag_list) == 1 and tag_list[0] == "no_activity":
        return health_icons["healthy"] + health_icons["no_activity"]

    # Loop through tags, get the symbol, and add it to the list
    for tag in tag_list:
        symbol_list.append(health_icons.get(tag, tag))
    return "".join(symbol_list)



"""MDQView Class: A View that computes various endpoint data quality metrics"""

from typing import Union
import pandas as pd

# Workbench Imports
from workbench.api import FeatureSet, Model, Endpoint
from workbench.core.views.view import View
from workbench.core.views.pandas_to_view import PandasToView
from workbench.algorithms.dataframe.row_tagger import RowTagger
from workbench.algorithms.dataframe.residuals_calculator import ResidualsCalculator


class MDQView:
    """MDQView Class: A View that computes various model/feature quality metrics

    Common Usage:
        ```python
        # Grab a FeatureSet and an Endpoint
        fs = FeatureSet("abalone_features")
        endpoint = Endpoint("abalone-regression")

        # Create a ModelDataQuality View
        mdq_view = MDQView.create(fs, endpoint=endpoint, id_column="id")
        my_df = mdq_view.pull_dataframe(limit=5)

        # Query the view
        df = mdq_view.query(f"SELECT * FROM {mdq_view.table} where residuals > 0.5")
        ```
    """

    @classmethod
    def create(
        cls,
        fs: FeatureSet,
        endpoint: Endpoint,
        id_column: str,
        use_reference_model: bool = False,
    ) -> Union[View, None]:
        """Create a Model Data Quality View with metrics

        Args:
            fs (FeatureSet): The FeatureSet object
            endpoint (Endpoint): The Endpoint object to use for the target and features
            id_column (str): The name of the id column (must be defined for join logic)
            use_reference_model (bool): Use the reference model for inference (default: False)

        Returns:
            Union[View, None]: The created View object (or None if failed)
        """
        # Log view creation
        fs.log.important("Creating Model Data Quality View...")

        # Get the target and feature columns from the endpoints model input
        model_input = Model(endpoint.get_input())
        target = model_input.target()
        features = model_input.features()

        # Pull in data from the source table
        df = fs.data_source.query(f"SELECT * FROM {fs.data_source.name}")

        # Check if the target and features are available in the data source
        missing_columns = [col for col in [target] + features if col not in df.columns]
        if missing_columns:
            fs.log.error(f"Missing columns in data source: {missing_columns}")
            return None

        # Check if the target is categorical
        categorical_target = not pd.api.types.is_numeric_dtype(df[target])

        # Compute row tags with RowTagger
        row_tagger = RowTagger(
            df,
            features=features,
            id_column=id_column,
            target_column=target,
            within_dist=0.25,
            min_target_diff=1.0,
            outlier_df=fs.data_source.outliers(),
            categorical_target=categorical_target,
        )
        mdq_df = row_tagger.tag_rows()

        # Rename and compute data quality scores based on tags
        mdq_df.rename(columns={"tags": "data_quality_tags"}, inplace=True)

        # We're going to compute a data_quality score based on the tags.
        mdq_df["data_quality"] = mdq_df["data_quality_tags"].apply(cls.calculate_data_quality)

        # Compute residuals using ResidualsCalculator
        if use_reference_model:
            residuals_calculator = ResidualsCalculator()
        else:
            residuals_calculator = ResidualsCalculator(endpoint=endpoint)
        residuals_df = residuals_calculator.fit_transform(df[features], df[target])

        # Add id_column to the residuals dataframe and merge with mdq_df
        residuals_df[id_column] = df[id_column]

        # Drop overlapping columns in mdq_df (except for the id_column) to avoid _x and _y suffixes
        overlap_columns = [col for col in residuals_df.columns if col in mdq_df.columns and col != id_column]
        mdq_df = mdq_df.drop(columns=overlap_columns)

        # Merge the DataFrames, with the id_column as the join key
        mdq_df = mdq_df.merge(residuals_df, on=id_column, how="left")

        # Delegate view creation to PandasToView
        view_name = "mdq_ref" if use_reference_model else "mdq"
        return PandasToView.create(view_name, fs, df=mdq_df, id_column=id_column)

    @staticmethod
    def calculate_data_quality(tags):
        score = 1.0  # Start with the default score
        if "coincident" in tags:
            score -= 1.0
        if "htg" in tags:
            score -= 0.5
        if "outlier" in tags:
            score -= 0.25
        score = max(0.0, score)
        return score


if __name__ == "__main__":
    """Exercise the MDQView functionality"""
    from workbench.api import FeatureSet, Endpoint

    # Get the FeatureSet
    my_fs = FeatureSet("abalone_features")

    # Grab the Endpoint
    my_endpoint = Endpoint("abalone-regression")

    # Create a MDQView
    mdq_view = MDQView.create(my_fs, endpoint=my_endpoint, id_column="id")

    # Pull the data quality dataframe
    my_df = mdq_view.pull_dataframe(limit=5)
    print(my_df)

    # Query the view
    df = mdq_view.query(f"SELECT * FROM {mdq_view.table} where residuals > 0.5")



"""Onboard a model both interactively and programmatically."""

import logging

# Workbench Imports
from workbench.api.model import Model, ModelType

# Setup logging
log = logging.getLogger("workbench")

# Grab one of the test models
test_model = Model("abalone-regression")

# Onboard the model interactively
test_model.onboard()

# Onboard the model programmatically
model_type = ModelType.REGRESSOR
target_column = "class_number_of_rings"
feature_list = ["length", "diameter", "height", "whole_weight", "shucked_weight", "viscera_weight", "shell_weight"]
endpoints = ["abalone-regression"]
owner = "workbench"
test_model.onboard_with_args(model_type, target_column, feature_list, endpoints, owner)



"""FeatureSet: Manages AWS Feature Store/Group creation and management.
FeatureSets are set up so they can easily be queried with AWS Athena.
All FeatureSets are run through a full set of Exploratory Data Analysis (EDA)
techniques (data quality, distributions, stats, outliers, etc.) FeatureSets
can be viewed and explored within the Workbench Dashboard UI."""

from typing import Union
from pathlib import Path
import pandas as pd

# Workbench Imports
from workbench.core.artifacts.artifact import Artifact
from workbench.core.artifacts.feature_set_core import FeatureSetCore
from workbench.core.transforms.features_to_model.features_to_model import FeaturesToModel
from workbench.api.model import Model, ModelType


class FeatureSet(FeatureSetCore):
    """FeatureSet: Workbench FeatureSet API Class

    Common Usage:
        ```python
        my_features = FeatureSet(name)
        my_features.details()
        my_features.to_model(
            name="abalone-regression",
            model_type=ModelType.REGRESSOR,
            target_column="class_number_of_rings"
            feature_list=["my", "best", "features"])
        )
        ```
    """

    def details(self, **kwargs) -> dict:
        """FeatureSet Details

        Returns:
            dict: A dictionary of details about the FeatureSet
        """
        return super().details(**kwargs)

    def query(self, query: str, **kwargs) -> pd.DataFrame:
        """Query the AthenaSource

        Args:
            query (str): The query to run against the FeatureSet

        Returns:
            pd.DataFrame: The results of the query
        """
        return super().query(query, **kwargs)

    def pull_dataframe(self, limit: int = 50000, include_aws_columns=False) -> pd.DataFrame:
        """Return a DataFrame of ALL the data from this FeatureSet

        Args:
            limit (int): Limit the number of rows returned (default: 50000)
            include_aws_columns (bool): Include the AWS columns in the DataFrame (default: False)

        Returns:
            pd.DataFrame: A DataFrame of ALL the data from this FeatureSet

        Note:
            Obviously this is not recommended for large datasets :)
        """

        # Get the table associated with the data
        self.log.info(f"Pulling data from {self.name}...")
        pull_query = f'SELECT * FROM "{self.athena_table}" LIMIT {limit}'
        df = self.query(pull_query)

        # Drop any columns generated from AWS
        if not include_aws_columns:
            aws_cols = ["write_time", "api_invocation_time", "is_deleted", "event_time"]
            df = df.drop(columns=aws_cols, errors="ignore")
        return df

    def to_model(
        self,
        name: str,
        model_type: ModelType,
        tags: list = None,
        description: str = None,
        feature_list: list = None,
        target_column: str = None,
        model_class: str = None,
        model_import_str: str = None,
        custom_script: Union[str, Path] = None,
        custom_args: dict = None,
        training_image: str = "xgb_training",
        inference_image: str = "xgb_inference",
        inference_arch: str = "x86_64",
        **kwargs,
    ) -> Union[Model, None]:
        """Create a Model from the FeatureSet

        Args:

            name (str): The name of the Model to create
            model_type (ModelType): The type of model to create (See workbench.model.ModelType)
            tags (list, optional): Set the tags for the model.  If not given tags will be generated.
            description (str, optional): Set the description for the model. If not give a description is generated.
            feature_list (list, optional): Set the feature list for the model. If not given a feature list is generated.
            target_column (str, optional): The target column for the model (use None for unsupervised model)
            model_class (str, optional): Model class to use (e.g. "KMeans", "PyTorch", default: None)
            model_import_str (str, optional): The import for the model (e.g. "from sklearn.cluster import KMeans")
            custom_script (str, optional): The custom script to use for the model (default: None)
            training_image (str, optional): The training image to use (default: "xgb_training")
            inference_image (str, optional): The inference image to use (default: "xgb_inference")
            inference_arch (str, optional): The architecture to use for inference (default: "x86_64")
            kwargs (dict, optional): Additional keyword arguments to pass to the model

        Returns:
            Model: The Model created from the FeatureSet (or None if the Model could not be created)
        """

        # Ensure the model_name is valid
        if name:
            if not Artifact.is_name_valid(name, delimiter="-", lower_case=False):
                self.log.critical(f"Invalid Model name: {name}, not creating Model!")
                return None

        # If the model_name wasn't given generate it
        else:
            name = self.name.replace("_features", "") + "-model"
            name = Artifact.generate_valid_name(name, delimiter="-")

        # Create the Model Tags
        tags = [name] if tags is None else tags

        # If the model_class is PyTorch, ensure we set the training and inference images
        if model_class and model_class.lower() == "pytorch":
            training_image = "pytorch_training"
            inference_image = "pytorch_inference"

        # Transform the FeatureSet into a Model
        features_to_model = FeaturesToModel(
            feature_name=self.name,
            model_name=name,
            model_type=model_type,
            model_class=model_class,
            model_import_str=model_import_str,
            custom_script=custom_script,
            custom_args=custom_args,
            training_image=training_image,
            inference_image=inference_image,
            inference_arch=inference_arch,
        )
        features_to_model.set_output_tags(tags)
        features_to_model.transform(
            target_column=target_column, description=description, feature_list=feature_list, **kwargs
        )

        # Return the Model
        return Model(name)


if __name__ == "__main__":
    """Exercise the FeatureSet Class"""
    from pprint import pprint

    # Retrieve an existing FeatureSet
    my_features = FeatureSet("test_features")
    pprint(my_features.summary())
    pprint(my_features.details())

    # Create a Model from the FeatureSet
    my_model = my_features.to_model(name="test-model", model_type=ModelType.REGRESSOR, target_column="iq_score")



import argparse
import redis
import heapq


def redis_reaper(host, port, size_limit=512000, expire=7):
    """Reap keys from Redis based on size and age.

    Args:
        host (str): Redis host.
        port (int): Redis port.
        size_limit (int): Size limit in bytes (default: 500KB).
        expire (int): Number of days before a key is considered expired (default: 7 days).
    """
    # Connect to Redis
    client = redis.Redis(host=host, port=port)
    total_keys = client.dbsize()
    keys_deleted = 0

    # Use a min-heap to track the 5 largest sizes
    largest_keys = []

    # Iterate through all keys
    for key in client.scan_iter():
        # Check the size of the value
        size = client.memory_usage(key) or 0

        # Track the largest keys
        heapq.heappush(largest_keys, (size, key))
        if len(largest_keys) > 5:
            heapq.heappop(largest_keys)

        if size > size_limit:
            print(f"Deleting key {key} (size: {size} bytes)")
            client.delete(key)
            keys_deleted += 1
            continue

        # Check if the key has a last modified timestamp
        key_info = client.object("idletime", key)
        if key_info is not None:
            key_age_days = key_info / (60 * 60 * 24)
            if key_age_days > expire:
                print(f"Deleting key {key} (age: {key_age_days:.2f} days)")
                client.delete(key)
                keys_deleted += 1

        # Check if the key has the word "Crashed" in its name
        if b"Crashed" in key:
            print(f"Deleting key {key} (contains 'Crashed')")
            client.delete(key)
            keys_deleted += 1

    # Report
    print(f"\nTotal keys: {total_keys}")
    print("\nTop 5 largest keys (by size):")
    for size, key in sorted(largest_keys, reverse=True):
        print(f"Key: {key}, Size: {size} bytes")

    print(f"\nTotal keys deleted: {keys_deleted}")


if __name__ == "__main__":
    # Argument parsing
    parser = argparse.ArgumentParser(description="Redis Reaper: Delete keys based on size and age.")
    parser.add_argument("--host", type=str, default="localhost", help="Redis host")
    parser.add_argument("-p", "--port", type=int, default=6379, help="Redis port (default: 6379)")
    parser.add_argument("--size-limit", type=int, default=1048576, help="Size limit in bytes (default: 1MB)")
    parser.add_argument("--expire", type=int, default=7, help="Expiration days (default: 7 days)")
    args = parser.parse_args()

    # Run the Redis Reaper
    redis_reaper(args.host, args.port, args.size_limit, args.expire)



import dash
import dash_bootstrap_components as dbc
from dash import dcc, html, Input, Output

# Initialize the Dash app with a Bootstrap theme
app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

# Layout of the app
app.layout = html.Div(
    [
        dbc.Button("Switch to Dark Theme", id="theme-button", color="primary"),
        dcc.Graph(
            id="scatter-plot",
            figure={
                "data": [
                    {"x": [1, 2, 3], "y": [4, 1, 2], "type": "scatter", "name": "SF"},
                ],
                "layout": {
                    "title": "Scatter Plot Example",
                    "plot_bgcolor": "#ffffff",
                    "paper_bgcolor": "#ffffff",
                    "font": {"color": "#000000"},
                },
            },
        ),
    ]
)


# Callback to switch themes dynamically
@app.callback(Output("scatter-plot", "figure"), [Input("theme-button", "n_clicks")], prevent_initial_call=True)
def switch_theme(n_clicks):
    dark_mode = n_clicks % 2 != 0
    if dark_mode:
        return {
            "data": [{"x": [1, 2, 3], "y": [4, 1, 2], "type": "scatter", "name": "SF"}],
            "layout": {
                "title": "Scatter Plot Example",
                "plot_bgcolor": "#333333",
                "paper_bgcolor": "#333333",
                "font": {"color": "#ffffff"},
            },
        }
    else:
        return {
            "data": [{"x": [1, 2, 3], "y": [4, 1, 2], "type": "scatter", "name": "SF"}],
            "layout": {
                "title": "Scatter Plot Example",
                "plot_bgcolor": "#ffffff",
                "paper_bgcolor": "#ffffff",
                "font": {"color": "#000000"},
            },
        }


# Run the app
if __name__ == "__main__":
    app.run(debug=True)



"""Endpoint: Manages AWS Endpoint creation and deployment.
Endpoints are automatically set up and provisioned for deployment into AWS.
Endpoints can be viewed in the AWS Sagemaker interfaces or in the Workbench
Dashboard UI, which provides additional model details and performance metrics"""

import pandas as pd

# Workbench Imports
from workbench.core.artifacts.endpoint_core import EndpointCore


class Endpoint(EndpointCore):
    """Endpoint: Workbench Endpoint API Class

    Common Usage:
        ```python
        my_endpoint = Endpoint(name)
        my_endpoint.details()
        my_endpoint.inference(eval_df)
        ```
    """

    def details(self, **kwargs) -> dict:
        """Endpoint Details

        Returns:
            dict: A dictionary of details about the Endpoint
        """
        return super().details(**kwargs)

    def inference(
        self, eval_df: pd.DataFrame, capture_name: str = None, id_column: str = None, drop_error_rows: bool = False
    ) -> pd.DataFrame:
        """Run inference on the Endpoint using the provided DataFrame

        Args:
            eval_df (pd.DataFrame): The DataFrame to run predictions on
            capture_name (str, optional): The Name of the capture to use (default: None)
            id_column (str, optional): The name of the column to use as the ID (default: None)
            drop_error_rows (bool): Whether to drop rows with errors (default: False)

        Returns:
            pd.DataFrame: The DataFrame with predictions
        """
        return super().inference(eval_df, capture_name, id_column, drop_error_rows)

    def auto_inference(self, capture: bool = False) -> pd.DataFrame:
        """Run inference on the Endpoint using the FeatureSet evaluation data

        Args:
            capture (bool): Capture the inference results

        Returns:
            pd.DataFrame: The DataFrame with predictions
        """
        return super().auto_inference(capture)

    def fast_inference(self, eval_df: pd.DataFrame, threads: int = 4) -> pd.DataFrame:
        """Run inference on the Endpoint using the provided DataFrame

        Args:
            eval_df (pd.DataFrame): The DataFrame to run predictions on
            threads (int): The number of threads to use (default: 4)

        Returns:
            pd.DataFrame: The DataFrame with predictions

        Note:
            There's no sanity checks or error handling... just FAST Inference!
        """
        return super().fast_inference(eval_df, threads=threads)

    def cross_fold_inference(self, nfolds: int = 5) -> dict:
        """Run cross-fold inference (only works for XGBoost models)

        Args:
            nfolds (int): The number of folds to use for cross-validation (default: 5)

        Returns:
            dict: A dictionary with fold results
        """
        return super().cross_fold_inference(nfolds)


if __name__ == "__main__":
    """Exercise the Endpoint Class"""
    from workbench.api.model import Model
    from workbench.api.feature_set import FeatureSet
    from pprint import pprint

    # Retrieve an existing Data Source
    my_endpoint = Endpoint("abalone-regression")
    pprint(my_endpoint.summary())

    # Run predictions on the Endpoint
    model = Model(my_endpoint.get_input())
    my_features = FeatureSet(model.get_input())
    table = my_features.view("training").table
    df = my_features.query(f'SELECT * FROM "{table}" where training = FALSE')
    results = my_endpoint.inference(df)
    target = model.target()
    pprint(results[[target, "prediction"]])

    # Run predictions using the auto_inference method
    auto_results = my_endpoint.auto_inference()
    pprint(auto_results[[target, "prediction"]])

    # Run predictions using the fast_inference method
    fast_results = my_endpoint.fast_inference(df)



from workbench.api.graph_store import GraphStore
from workbench.utils.graph_utils import modified, details, get_tags


def test_general_info():
    """Simple tests of Graph functionality"""
    graph_store = GraphStore()

    # Call the various methods
    assert graph_store.check("test/karate_club")

    test_graph = graph_store.get("test/karate_club")

    # Modification Time
    print(modified(test_graph))

    # Get the tags associated with this Graph
    print(f"Tags: {get_tags(test_graph)}")

    # Details
    print(details(test_graph))


if __name__ == "__main__":

    # Run the tests
    test_general_info()

    print("All tests passed!")



"""This Script creates the 'Training Adjusted' Artifacts in AWS needed for the tests

FeatureSets:
    - Create a training view for abalone_features
Models:
    - abalone-regression-100
Endpoints:
    - abalone-regression-end-100
"""

import sys
import time
import logging
from pathlib import Path
from workbench.core.artifacts.feature_set_core import FeatureSetCore
from workbench.core.artifacts.model_core import ModelCore, ModelType
from workbench.core.artifacts.endpoint_core import EndpointCore
from workbench.core.transforms.features_to_model.features_to_model import FeaturesToModel
from workbench.core.transforms.model_to_endpoint.model_to_endpoint import ModelToEndpoint

# Setup the logger
log = logging.getLogger("workbench")


if __name__ == "__main__":

    # Get the path to the dataset in the repository data directory
    abalone_data_path = Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "abalone.csv"

    # Recreate Flag in case you want to recreate the artifacts
    recreate = False

    # Create a training view of the test_features
    log.important("Creating training view for abalone_features...")
    fs = FeatureSetCore("abalone_features")
    fs.set_training_holdouts("id", holdout_ids=range(100))  # Just the first 100 ids

    # Create the abalone_regression Model
    if recreate or not ModelCore("abalone-regression-100").exists():
        features_to_model = FeaturesToModel(
            "abalone_features", "abalone-regression-100", model_type=ModelType.REGRESSOR
        )
        features_to_model.set_output_tags(["abalone", "regression"])
        features_to_model.transform(
            target_column="class_number_of_rings", description="Abalone Regression Model", train_all_data=True
        )
        log.info("Waiting for the Model to be created...")
        time.sleep(10)

    # Create the abalone_regression Endpoint
    if recreate or not EndpointCore("abalone-regression-end-100").exists():
        model_to_endpoint = ModelToEndpoint("abalone-regression-100", "abalone-regression-end-100")
        model_to_endpoint.set_output_tags(["abalone", "regression"])
        model_to_endpoint.transform()



"""DataToFeaturesChunk: Class to Transform a DataSource into a FeatureSet using Chunking"""

import awswrangler as wr

# Local imports
from workbench.core.transforms.transform import Transform
from workbench.core.transforms.pandas_transforms.pandas_to_features_chunked import (
    PandasToFeaturesChunked,
)
from workbench.core.artifacts.data_source_factory import DataSourceFactory


class DataToFeaturesChunk(Transform):
    """DataToFeaturesChunk: Class to Transform a DataSource into a FeatureSet using Chunking

    Common Usage:
        ```python
        data_to_features = DataToFeaturesChunk(input_name, output_name, 50000)
        data_to_features.set_output_tags(["heavy", "whatever"])
        data_to_features.transform(query, id_column, event_time_column=None)
        ```
    """

    def __init__(self, input_name: str, output_name: str, chunk_size: int = 50000):
        """DataToFeaturesChunk Initialization"""

        # Call superclass init
        super().__init__(input_name, output_name)

        # Set up all my instance attributes
        self.id_column = None
        self.event_time_column = None
        self.chunk_size = chunk_size
        self.input_data_source = DataSourceFactory(input_name)
        self.ds_database = "workbench"
        self.cat_column_info = {}
        self.chunked_to_features = None

    def set_categorical_info(self, cat_column_info: dict[list[str]]):
        """Set the Categorical Columns Information
        Args:
            cat_column_info (dict[list[str]]): Dictionary of categorical columns and their possible values
        """
        self.cat_column_info = cat_column_info

    def detect_categoral_columns(self):
        """Automatically figure out which columns are categorical"""
        self.log.info("Automatically computing the categorical columns")

        # DataSources will compute value_counts for each object/str column
        value_counts = self.input_data_source.value_counts()
        for column, value_info in value_counts.items():
            # Hack to avoid IP columns
            if column.endswith("ip"):
                self.log.info(f"Skipping column {column}...")
                continue
            # How many unique values are there?
            unique = len(value_info.keys())
            if 1 < unique < 6:
                self.log.info(f"Column {column} is categorical (unique={unique})")
                self.cat_column_info[column] = list(value_info.keys())

    def pre_transform(self, **kwargs):
        """Figure out which fields are categorical"""

        # If the user didn't specify any categorical columns, try to figure them out
        if not self.cat_column_info:
            self.detect_categoral_columns()

    def transform_impl(self, query, id_column: str, event_time_column: str = None):
        """Convert the Data Source into a Feature Set using Chunking"""

        # Create our PandasToFeaturesChunked class
        self.chunked_to_features = PandasToFeaturesChunked(self.output_name, id_column, event_time_column)
        self.chunked_to_features.set_output_tags(self.output_tags)
        self.chunked_to_features.set_categorical_info(self.cat_column_info)
        self.chunked_to_features.pre_transform()

        # Read in the data from Athena in chunks
        for chunk in wr.athena.read_sql_query(
            query,
            database=self.ds_database,
            ctas_approach=False,
            chunksize=self.chunk_size,
            boto3_session=self.boto3_session,
        ):
            # Hand off each chunk of data
            self.chunked_to_features.add_chunk(chunk)

    def post_transform(self, **kwargs):
        """Post-Transform: Any Post Transform Steps"""
        self.log.info("Post-Transform: Completing FeatureSet Offline Storage...")
        self.chunked_to_features.post_transform()


if __name__ == "__main__":
    """Exercise the DataToFeaturesChunk Class"""

    # Create my DF to Feature Set Transform
    data_to_features = DataToFeaturesChunk("http_10795", "http_features")
    data_to_features.set_output_tags(["http", "nomic"])

    # Store this dataframe as a Workbench Feature Set

    # Old Stuff
    """
    fields = [
        "timestamp",
        "flow_id",
        "in_iface",
        "proto",
        "dns_type",
        "dns_rrtype",
        "dns_flags",
        "dns_rcode",
    ]
    query = f"SELECT {', '.join(fields)} FROM heavy_dns limit 1000"
    """
    query = "SELECT * FROM http_10795"
    data_to_features.transform(query=query, id_column="flow_id", event_time_column="timestamp")



"""DFStore: Fast/efficient storage of DataFrames using AWS S3/Parquet/Snappy"""

from datetime import datetime
from typing import Union
import logging
import pandas as pd

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_df_store import AWSDFStore


class DFStore(AWSDFStore):
    """DFStore: Fast/efficient storage of DataFrames using AWS S3/Parquet/Snappy

    Common Usage:
        ```python
        df_store = DFStore()

        # List Data
        df_store.list()

        # Add DataFrame
        df = pd.DataFrame({"A": [1, 2], "B": [3, 4]})
        df_store.upsert("/test/my_data", df)

        # Retrieve DataFrame
        df = df_store.get("/test/my_data")
        print(df)

        # Delete Data
        df_store.delete("/test/my_data")
        ```
    """

    def __init__(self, path_prefix: Union[str, None] = None):
        """DFStore Init Method

        Args:
            path_prefix (Union[str, None], optional): Add a path prefix to storage locations (Defaults to None)
        """
        self.log = logging.getLogger("workbench")

        # Initialize the SuperClass
        super().__init__(path_prefix=path_prefix)

    def list(self, include_cache: bool = False) -> list:
        """List all the objects in the data_store prefix.

        Args:
            include_cache (bool, optional): Include cache objects in the list (Defaults to False).

        Returns:
            list: A list of all the objects in the data_store prefix.
        """
        return super().list(include_cache=include_cache)

    def summary(self, include_cache: bool = False) -> pd.DataFrame:
        """Return a nicely formatted summary of object locations, sizes (in MB), and modified dates.

        Args:
            include_cache (bool, optional): Include cache objects in the summary (Defaults to False).

        Returns:
            pd.DataFrame: A formatted DataFrame with the summary details.
        """
        return super().summary(include_cache=include_cache)

    def details(self, include_cache: bool = False) -> pd.DataFrame:
        """Return a DataFrame with detailed metadata for all objects in the data_store prefix.

        Args:
            include_cache (bool, optional): Include cache objects in the details (Defaults to False).

        Returns:
            pd.DataFrame: A DataFrame with detailed metadata for all objects in the data_store prefix.
        """
        return super().details(include_cache=include_cache)

    def check(self, location: str) -> bool:
        """Check if a DataFrame exists at the specified location

        Args:
            location (str): The location of the data to check.

        Returns:
            bool: True if the data exists, False otherwise.
        """
        return super().check(location)

    def get(self, location: str) -> Union[pd.DataFrame, None]:
        """Retrieve a DataFrame from AWS S3.

        Args:
            location (str): The location of the data to retrieve.

        Returns:
            pd.DataFrame: The retrieved DataFrame or None if not found.
        """
        _df = super().get(location)
        if _df is None:
            self.log.error(f"Dataframe not found at location: {location}")
        return _df

    def upsert(self, location: str, data: Union[pd.DataFrame, pd.Series]):
        """Insert or update a DataFrame or Series in the AWS S3.

        Args:
            location (str): The location of the data.
            data (Union[pd.DataFrame, pd.Series]): The data to be stored.
        """
        super().upsert(location, data)

    def last_modified(self, location: str) -> Union[datetime, None]:
        """Get the last modified date of the DataFrame at the specified location.

        Args:
            location (str): The location of the data to check.

        Returns:
            Union[datetime, None]: The last modified date of the DataFrame or None if not found.
        """
        return super().last_modified(location)

    def delete(self, location: str):
        """Delete a DataFrame from the AWS S3.

        Args:
            location (str): The location of the data to delete.
        """
        super().delete(location)


if __name__ == "__main__":
    """Exercise the DFStore Class"""
    import time

    # Create a DFStore manager
    df_store = DFStore()

    # Details of the Dataframe Store
    print("Detailed Data...")
    print(df_store.details())

    # Add a new DataFrame
    my_df = pd.DataFrame({"A": [1, 2], "B": [3, 4]})
    df_store.upsert("/testing/test_data", my_df)

    # Get the DataFrame
    print(f"Getting data 'test_data':\n{df_store.get('/testing/test_data')}")

    # Now let's test adding a Series
    series = pd.Series([1, 2, 3, 4], name="Series")
    df_store.upsert("/testing/test_series", series)
    print(f"Getting data 'test_series':\n{df_store.get('/testing/test_series')}")

    # Summary of the data
    print("Summary Data...")
    print(df_store.summary())

    # Repr of the DFStore object
    print("DFStore Object:")
    print(df_store)

    # Check if the data exists
    print("Check if data exists...")
    print(df_store.check("/testing/test_data"))
    print(df_store.check("/testing/test_series"))

    # Time the check
    start_time = time.time()
    print(df_store.check("/testing/test_data"))
    print("--- Check %s seconds ---" % (time.time() - start_time))

    # Now delete the test data
    df_store.delete("/testing/test_data")
    df_store.delete("/testing/test_series")

    # Check if the data exists
    print("Check if data exists...")
    print(df_store.check("/testing/test_data"))
    print(df_store.check("/testing/test_series"))

    # Add a bunch of dataframes and then test recursive delete
    for i in range(10):
        df_store.upsert(f"/testing/data_{i}", pd.DataFrame({"A": [1, 2], "B": [3, 4]}))
    print("Before Recursive Delete:")
    print(df_store.summary())
    df_store.delete_recursive("/testing")
    print("After Recursive Delete:")
    print(df_store.summary())

    # Get a non-existent DataFrame
    print("Getting non-existent data...")
    print(df_store.get("/testing/no_where"))



"""Welcome to the Workbench API Classes

These class provide high-level APIs for the Workbench package, offering easy access to its core classes:

- DataSource: Manages AWS Data Catalog and Athena
- FeatureSet: Manages AWS Feature Store and Feature Groups
- Model: Manages the training and deployment of AWS Model Groups and Packages
- ModelType: Enum for the different model types supported by Workbench
- Endpoint: Manages the deployment and invocations/inference on AWS Endpoints
- Meta: Provides an API to retrieve AWS Metadata for the above classes
- ParameterStore: Manages AWS Parameter Store
- DFStore: Manages DataFrames in AWS S3
"""

from .data_source import DataSource
from .feature_set import FeatureSet
from .model import Model, ModelType
from .endpoint import Endpoint
from .meta import Meta
from .parameter_store import ParameterStore
from .df_store import DFStore

__all__ = [
    "DataSource",
    "FeatureSet",
    "Model",
    "ModelType",
    "Endpoint",
    "Meta",
    "ParameterStore",
    "DFStore",
]



import json
import boto3
import random
import string
import gzip
import os
from datetime import datetime, timezone

# Workbench Imports
from workbench.utils.datetime_utils import datetime_to_iso8601

# Set up S3 client
s3 = boto3.client("s3")

# Set up some variables
num_files = 10  # Number of files to create
num_records_per_file = 1000  # Number of records to create in each file


# Define a function to generate random data
def generate_data():
    return {
        "id": "".join(random.choices(string.ascii_letters + string.digits, k=8)),
        "name": "".join(random.choices(string.ascii_letters, k=10)),
        "age": random.randint(18, 65),
        "address": "".join(random.choices(string.ascii_letters + string.digits + ", .", k=20)),
        "date": datetime_to_iso8601(datetime.now(timezone.utc)),
    }


# Loop through and generate the files
for i in range(num_files):
    # Set up the file name
    filename = f"data_{i}.jsonl"
    print(filename)

    # Generate the records
    records = []
    for j in range(num_records_per_file):
        records.append(generate_data())

    # Write the records to the file in JSONL format
    with open(filename, "w") as f:
        for record in records:
            f.write(json.dumps(record) + "\n")

    # Compress the file and save as gzipped file
    with open(filename, "rb") as f_in:
        with gzip.open(f"{filename}.gz", "wb") as f_out:
            f_out.writelines(f_in)

    # Upload the file to S3
    s3_file_path = f"incoming-data/jsonl-data/{filename}.gz"
    s3.upload_file(f"{filename}.gz", "scp-workbench-artifacts", s3_file_path)

    # Delete the local copy of the file
    os.remove(filename)
    os.remove(f"{filename}.gz")

print("Done!")



"""PageView: Pulls from the Cloud Metadata and performs page specific data processing"""

from abc import ABC, abstractmethod

import logging


class PageView(ABC):
    def __init__(self):
        """PageView: Pulls from the Cloud Metadata and performs page specific data processing"""
        self.log = logging.getLogger("workbench")

    @abstractmethod
    def refresh(self):
        """Refresh the data associated with this page view"""
        pass



"""Script that loops through all models and checks if they are ready"""

import logging

# Workbench Imports
from workbench.api import Meta, Model

# Setup logging
log = logging.getLogger("workbench")

# Get all the models
models = Meta().models()
for model_name in models["Model Group"]:
    m = Model(model_name)
    if m.ready():
        log.important(f"Model {model_name} is ready!")
    else:
        log.important(f"Model {model_name} is not ready...Calling onboard.... ")
        m.onboard()



import sagemaker
import logging
from botocore.exceptions import ClientError

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Hardcoded values
MODEL_PACKAGE_GROUP_NAME = "abalone-regression"
ENDPOINT_NAME = f"{MODEL_PACKAGE_GROUP_NAME}"
MEM_SIZE = 4096
MAX_CONCURRENCY = 10
ROLE_ARN = "arn:aws:iam::507740646243:role/Workbench-ExecutionRole"

# Get a standard SageMaker session
session = sagemaker.Session()


# Get the latest approved model package
def get_latest_model_package(model_package_group_name):
    model_packages = session.sagemaker_client.list_model_packages(
        ModelPackageGroupName=model_package_group_name,
        ModelApprovalStatus="Approved",
        SortBy="CreationTime",
        SortOrder="Descending",
    )
    if not model_packages["ModelPackageSummaryList"]:
        raise ValueError(f"No approved model packages found in group: {model_package_group_name}")

    latest_model_package_arn = model_packages["ModelPackageSummaryList"][0]["ModelPackageArn"]

    # Describe the model package to get details
    model_package_details = session.sagemaker_client.describe_model_package(ModelPackageName=latest_model_package_arn)

    logger.info(f"Model Package Details: {model_package_details}")
    return latest_model_package_arn


# Delete existing model if it exists
def delete_model_if_exists(model_name):
    try:
        session.sagemaker_client.delete_model(ModelName=model_name)
        logger.info(f"Deleted existing model: {model_name}")
    except ClientError as e:
        if (
            e.response["Error"]["Code"] == "ValidationException"
            and "Could not find model" in e.response["Error"]["Message"]
        ):
            logger.info(f"Model {model_name} does not exist, no need to delete")
        else:
            raise


# Create a serverless endpoint
def create_serverless_endpoint(model_package_arn, endpoint_name, role_arn, mem_size, max_concurrency):
    try:
        # Define the model name
        model_name = f"{endpoint_name}-model"

        # Delete the existing model if it exists
        delete_model_if_exists(model_name)

        # Create the model from the model package
        logger.info(f"Creating model with ARN: {model_package_arn}")
        session.sagemaker_client.create_model(
            ModelName=model_name,
            PrimaryContainer={
                "ModelPackageName": model_package_arn,
            },
            ExecutionRoleArn=role_arn,
        )

        # Define the endpoint configuration name
        endpoint_config_name = f"{endpoint_name}-config"

        # Create the endpoint configuration
        logger.info(
            f"Creating endpoint configuration with memory size: {mem_size} and max concurrency: {max_concurrency}"
        )
        session.sagemaker_client.create_endpoint_config(
            EndpointConfigName=endpoint_config_name,
            ProductionVariants=[
                {
                    "ServerlessConfig": {"MemorySizeInMB": mem_size, "MaxConcurrency": max_concurrency},
                    "ModelName": model_name,
                    "VariantName": "AllTraffic",
                }
            ],
        )

        # Create the endpoint
        logger.info(f"Creating endpoint: {endpoint_name}")
        response = session.create_endpoint(
            endpoint_name=endpoint_name, config_name=endpoint_config_name, wait=True, live_logging=True
        )

        logger.info(f"Endpoint creation response: {response}")
        logger.info(f"Endpoint {endpoint_name} created successfully.")

        # Check the endpoint logs
        endpoint_logs = session.sagemaker_client.describe_endpoint(EndpointName=endpoint_name)
        logger.info(f"Endpoint Logs: {endpoint_logs}")

    except ClientError as e:
        logger.error(f"ClientError: {e.response['Error']['Code']}, {e.response['Error']['Message']}")
        if "LogResult" in e.response:
            log_result = e.response["LogResult"]
            logger.error(f"LogResult: {log_result}")
        raise
    except Exception as e:
        logger.error(f"Failed to create serverless endpoint: {str(e)}")
        raise


# Main execution
if __name__ == "__main__":
    try:
        latest_model_package_arn = get_latest_model_package(MODEL_PACKAGE_GROUP_NAME)
        logger.info(f"Latest model package ARN: {latest_model_package_arn}")
        create_serverless_endpoint(latest_model_package_arn, ENDPOINT_NAME, ROLE_ARN, MEM_SIZE, MAX_CONCURRENCY)
    except Exception as e:
        logger.error(f"Deployment failed: {str(e)}")
        raise



import dash
from dash import html, dcc, Input, Output, State
import pandas as pd
from dash_ag_grid import AgGrid

app = dash.Dash(__name__)

# Placeholder DataFrame
df = pd.DataFrame()

app.layout = html.Div(
    [
        html.Button("Update Table", id="update-button", n_clicks=0),
        AgGrid(
            id="my-table",
            columnSize="sizeToFit",
            dashGridOptions={
                "rowHeight": None,
                "domLayout": "normal",
                "rowSelection": "single",
                "filter": True,
            },
            style={"maxHeight": "200px", "overflow": "auto"},
        ),
        html.Div(id="selected-row-info"),  # Div to display selected row information
    ]
)


@app.callback(
    Output("my-table", "columnDefs"),
    Output("my-table", "rowData"),
    Output("my-table", "selectedRows"),
    Input("update-button", "n_clicks"),
    prevent_initial_call=True,
)
def update_data(n_clicks):
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    column_defs = [{"headerName": col, "field": col, "filter": "agTextColumnFilter"} for col in df.columns]
    row_data = df.to_dict("records")
    selected_rows = df.head(1).to_dict("records")
    selected_rows = {"ids": ["1"]}
    print("Updating Table")
    # return column_defs, row_data, selected_rows
    return (
        column_defs,
        row_data,
        selected_rows,
    )


@app.callback(Output("selected-row-info", "children"), Input("my-table", "selectedRows"))
def update_selected_row_info(selected_rows):
    selected_row_info = f"Selected Rows: {selected_rows}"
    print(selected_row_info)
    return selected_row_info


if __name__ == "__main__":
    app.run(debug=True)



"""PandasToFeaturesChunked: Class to manage a bunch of chunked Pandas DataFrames into a FeatureSet"""

import pandas as pd
from pandas.api.types import CategoricalDtype

# Local imports
from workbench.core.transforms.transform import Transform
from workbench.core.transforms.pandas_transforms.pandas_to_features import PandasToFeatures
from workbench.core.artifacts.feature_set_core import FeatureSetCore
from workbench.core.artifacts.artifact import Artifact


class PandasToFeaturesChunked(Transform):
    """PandasToFeaturesChunked:  Class to manage a bunch of chunked Pandas DataFrames into a FeatureSet

    Common Usage:
        ```python
        to_features = PandasToFeaturesChunked(output_name, id_column="id"/None, event_time_column="date"/None)
        to_features.set_output_tags(["abalone", "public", "whatever"])
        cat_column_info = {"sex": ["M", "F", "I"]}
        to_features.set_categorical_info(cat_column_info)
        to_features.add_chunk(df)
        to_features.add_chunk(df)
        ...
        to_features.finalize()
        ```
    """

    def __init__(self, output_name: str, id_column=None, event_time_column=None):
        """PandasToFeaturesChunked Initialization"""

        # Make sure the output_name is a valid name
        Artifact.is_name_valid(output_name)

        # Call superclass init
        super().__init__("DataFrame", output_name)

        # Set up all my instance attributes
        self.id_column = id_column
        self.event_time_column = event_time_column
        self.first_chunk = None
        self.pandas_to_features = PandasToFeatures(output_name)

    def set_categorical_info(self, cat_column_info: dict[list[str]]):
        """Set the Categorical Columns
        Args:
            cat_column_info (dict[list[str]]): Dictionary of categorical columns and their possible values
        """

        # Create the CategoricalDtypes
        cat_d_types = {}
        for col, vals in cat_column_info.items():
            cat_d_types[col] = CategoricalDtype(categories=vals)

        # Now set the CategoricalDtypes on our underlying PandasToFeatures
        self.pandas_to_features.categorical_dtypes = cat_d_types

    def add_chunk(self, chunk_df: pd.DataFrame):
        """Add a Chunk of Data to the FeatureSet"""

        # Is this the first chunk? If so we need to run the pre_transform
        if self.first_chunk is None:
            self.log.info(f"Adding first chunk {chunk_df.shape}...")
            self.first_chunk = chunk_df
            self.pandas_to_features.set_input(chunk_df, self.id_column, self.event_time_column)
            self.pandas_to_features.pre_transform()
            self.pandas_to_features.transform_impl()
        else:
            self.log.info(f"Adding chunk {chunk_df.shape}...")
            self.pandas_to_features.set_input(chunk_df, self.id_column, self.event_time_column)
            self.pandas_to_features.transform_impl()

    def pre_transform(self, **kwargs):
        """Pre-Transform: Create the Feature Group with Chunked Data"""

        # Loading data into a Feature Group takes a while, so set status to loading
        FeatureSetCore(self.output_name).set_status("loading")

    def transform_impl(self):
        """Required implementation of the Transform interface"""
        self.log.warning("PandasToFeaturesChunked.transform_impl() called.  This is a no-op.")

    def post_transform(self, **kwargs):
        """Post-Transform: Any Post Transform Steps"""
        self.pandas_to_features.post_transform()


if __name__ == "__main__":
    """Exercise the PandasToFeaturesChunked Class"""
    from workbench.api.data_source import DataSource

    # Load in a DataFrame from a DataSource and split it into chunks
    ds = DataSource("abalone_data")
    df = ds.query("select * from abalone_data")
    print(f"df.shape: {df.shape}")

    # Build DataFrames chunks of 1000 rows each
    chunk_size = 1000
    chunks = [df[i : i + chunk_size] for i in range(0, df.shape[0], chunk_size)]

    # Create our PandasToFeaturesChunked class
    to_features = PandasToFeaturesChunked("abalone_features")

    # Manually set the Categorical Columns
    categorical_column_info = {"sex": ["M", "F", "I"]}
    to_features.set_categorical_info(categorical_column_info)

    # Now loop through the chunks and add them to the FeatureSet
    for chunk in chunks:
        to_features.add_chunk(chunk)
    to_features.finalize()



import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import DBSCAN, HDBSCAN
from sklearn.cluster import MeanShift

# Workbench Imports
from workbench.api import FeatureSet, Model


# Sample code to run DBSCAN
def run_dbscan(df: pd.DataFrame, feature_list: list, eps: float = 0.5, min_samples: int = 5):
    """
    Run DBSCAN on the provided DataFrame using the specified feature list.

    Args:
        df (pd.DataFrame): The input DataFrame.
        feature_list (list): The list of features to use for clustering.
        eps (float): The maximum distance between two samples for them to be considered as in the same neighborhood.
        min_samples (int): The number of samples (or total weight) in a neighborhood for a point to be considered a core point.

    Returns:
        pd.Series: Cluster labels assigned by DBSCAN.
    """
    # Standardize the features
    scaler = StandardScaler()
    scaled_features = scaler.fit_transform(df[feature_list])

    # Run HDBSCAN
    db = HDBSCAN()
    df["db_cluster"] = db.fit_predict(scaled_features)
    return df


# Sample code to run Mean Shift
def run_mean_shift(df: pd.DataFrame, feature_list: list):
    """
    Run Mean Shift on the provided DataFrame using the specified feature list.

    Args:
        df (pd.DataFrame): The input DataFrame.
        feature_list (list): The list of features to use for clustering.

    Returns:
        pd.Series: Cluster labels assigned by Mean Shift.
    """
    # Standardize the features
    scaler = StandardScaler()
    scaled_features = scaler.fit_transform(df[feature_list])

    # Run Mean Shift
    mean_shift = MeanShift()
    df["mean_shift_cluster"] = mean_shift.fit_predict(scaled_features)
    return df


# Example usage
if __name__ == "__main__":
    # Grab the dataframe from the FeatureSet
    fs = FeatureSet("wine_features")
    df = fs.pull_dataframe()

    # Grab the Feature List from the Model
    model = Model("wine-classification")
    features = model.features()

    # Run DBSCAN and print the cluster labels
    df = run_dbscan(df, features)
    print(df)
    print(df["db_cluster"].value_counts())

    # Run Mean Shift and print the cluster labels
    df = run_mean_shift(df, features)
    print(df)
    print(df["mean_shift_cluster"].value_counts())

    # Grab the dataframe from the FeatureSet
    fs = FeatureSet("abalone_features")
    df = fs.pull_dataframe()

    # Grab the Feature List from the Model
    model = Model("abalone-regression")
    features = model.features()

    # Run DBSCAN and print the cluster labels
    df = run_dbscan(df, features)
    print(df)
    print(df["db_cluster"].value_counts())

    # Run Mean Shift and print the cluster labels
    df = run_mean_shift(df, features)
    print(df)
    print(df["mean_shift_cluster"].value_counts())



"""AWSAccountCheck runs a bunch of tests/checks to ensure Workbench AWS Setup"""

import sys
import logging
import awswrangler as wr
from botocore.exceptions import ClientError

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
from workbench.utils.config_manager import ConfigManager, FatalConfigError


class AWSAccountCheck:
    """AWSAccountCheck runs a bunch of tests/checks to ensure Workbench AWS Setup"""

    def __init__(self):
        """AWSAccountCheck Initialization"""
        self.log = logging.getLogger("workbench")

        # Create the AWSAccountClamp Class
        self.aws_clamp = AWSAccountClamp()

        # Grab our Workbench Bucket
        cm = ConfigManager()
        if not cm.config_okay():
            self.log.error("Workbench Configuration Incomplete...")
            self.log.error("Run the 'workbench' command and follow the prompts...")
            raise FatalConfigError()

        self.workbench_bucket = cm.get_config("WORKBENCH_BUCKET")

    def ensure_aws_catalog_db(self, catalog_db: str):
        """Ensure that the AWS Data Catalog Database exists"""
        self.log.important(f"Ensuring that the AWS Data Catalog Database {catalog_db} exists...")
        try:
            wr.catalog.create_database(catalog_db, exist_ok=True, boto3_session=self.aws_clamp.boto3_session)
        except ClientError as e:
            if e.response["Error"]["Code"] == "AccessDeniedException":
                self.log.error(f"AccessDeniedException {e}")
                self.log.error(f"Access denied while trying to create/access the catalog database '{catalog_db}'.")
                self.log.error("Create the database manually in the AWS Glue Console, or run this command:")
                self.log.error(f'aws glue create-database --database-input \'{{"Name": "{catalog_db}"}}\'')
                sys.exit(1)
            else:
                self.log.error(f"Unexpected error: {e}")
                sys.exit(1)

    def check(self):
        """Check if the AWS Account is set up Correctly"""
        self.log.info("*** Caller/Base AWS Identity Check ***")
        self.aws_clamp.check_aws_identity()
        print("\n")

        self.log.info("*** AWS Assumed Role Check ***")
        self.aws_clamp.check_assumed_role()
        print("\n")

        self.log.info("*** AWS Workbench Bucket Check ***")
        self.aws_clamp.check_workbench_bucket()
        print("\n")

        self.log.info("*** AWS Sagemaker Session/Client Check ***")
        sm_client = self.aws_clamp.sagemaker_client()
        for feature_group in sm_client.list_feature_groups()["FeatureGroupSummaries"]:
            self.log.info(str(feature_group))
        print("\n")

        # Check that the Glue Databases exist
        self.log.info("*** AWS Glue Databases Check ***")
        for catalog_db in ["workbench", "sagemaker_featurestore", "inference_store"]:
            self.ensure_aws_catalog_db(catalog_db)
        print("\n")

        self.log.info("AWS Account Clamp: AOK!")


if __name__ == "__main__":
    """Exercise the AWS Account Clamp Class"""

    # Create the class
    account_check = AWSAccountCheck()

    # Check that out AWS Account Clamp is working AOK
    account_check.check()



"""CachedDataSource: Caches the method results for Workbench DataSources"""

from typing import Union

import pandas as pd

# Workbench Imports
from workbench.core.artifacts.athena_source import AthenaSource
from workbench.core.artifacts.cached_artifact_mixin import CachedArtifactMixin


class CachedDataSource(CachedArtifactMixin, AthenaSource):
    """CachedDataSource: Caches the method results for Workbench DataSources

    Note: Cached method values may lag underlying DataSource changes.

    Common Usage:
        ```python
        my_data = CachedDataSource(name)
        my_data.details()
        my_data.health_check()
        my_data.workbench_meta()
        ```
    """

    def __init__(self, data_name: str, database: str = "workbench"):
        """CachedDataSource Initialization"""
        AthenaSource.__init__(self, data_name=data_name, database=database, use_cached_meta=True)

    @CachedArtifactMixin.cache_result
    def summary(self, **kwargs) -> dict:
        """Retrieve the DataSource Details.

        Returns:
            dict: A dictionary of details about the DataSource
        """
        return super().summary(**kwargs)

    @CachedArtifactMixin.cache_result
    def details(self, **kwargs) -> dict:
        """Retrieve the DataSource Details.

        Returns:
            dict: A dictionary of details about the DataSource
        """
        return super().details(**kwargs)

    @CachedArtifactMixin.cache_result
    def health_check(self, **kwargs) -> dict:
        """Retrieve the DataSource Health Check.

        Returns:
            dict: A dictionary of health check details for the DataSource
        """
        return super().health_check(**kwargs)

    @CachedArtifactMixin.cache_result
    def workbench_meta(self) -> Union[dict, None]:
        """Retrieve the Workbench Metadata for this DataSource.

        Returns:
            Union[dict, None]: Dictionary of Workbench metadata for this Artifact
        """
        return super().workbench_meta()

    def smart_sample(self) -> pd.DataFrame:
        """Retrieve the Smart Sample for this DataSource.

        Returns:
            pd.DataFrame: The Smart Sample DataFrame
        """
        return super().smart_sample()


if __name__ == "__main__":
    """Exercise the CachedDataSource Class"""
    from pprint import pprint

    # Retrieve an existing DataSource
    my_data = CachedDataSource("abalone_data")
    pprint(my_data.summary())
    pprint(my_data.details())
    pprint(my_data.health_check())
    pprint(my_data.workbench_meta())

    # Shutdown the ThreadPoolExecutor (note: users should NOT call this)
    my_data._shutdown()



import json
import os

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp


class ECSInfo:
    def __init__(self):
        # Fetch the ECS Metadata from a file that's mounted into the container
        self.ecs_metadata = self._fetch_ecs_metadata()

        # Spin up our AWSAccountClamp
        self.aws_account_clamp = AWSAccountClamp()
        self.boto3_session = self.aws_account_clamp.boto3_session
        self.ecs_client = self.boto3_session.client("ecs")
        self.elbv2_client = self.boto3_session.client("elbv2")

    def get_cluster_name(self):
        return self.ecs_metadata.get("Cluster", "Unknown")

    def get_service_name(self):
        containers = self.ecs_metadata.get("Containers", [])
        if containers:
            return containers[0].get("Labels", {}).get("com.amazonaws.ecs.service-name", "Unknown")
        return "Unknown"

    def _fetch_ecs_metadata(self):
        metadata_file_path = os.environ.get("ECS_CONTAINER_METADATA_FILE", "/etc/ecs_container_metadata.json")
        try:
            with open(metadata_file_path, "r") as f:
                metadata = json.load(f)
            return metadata
        except FileNotFoundError:
            print("Metadata file not found. Are you sure you're running this on ECS?")
            return {}

    def get_load_balancer_name(self):
        try:
            response = self.ecs_client.describe_services(
                cluster=self.get_cluster_name(), services=[self.get_service_name()]
            )

            # Extract Target Group ARN
            target_group_arn = None
            if "services" in response and len(response["services"]) > 0:
                if "loadBalancers" in response["services"][0] and len(response["services"][0]["loadBalancers"]) > 0:
                    target_group_arn = response["services"][0]["loadBalancers"][0]["targetGroupArn"]

            # Describe Load Balancers associated with the Target Group
            if target_group_arn:
                response = self.elbv2_client.describe_load_balancers()
                for lb in response["LoadBalancers"]:
                    load_balancer_arn = lb["LoadBalancerArn"]
                    load_balancer_name = "/".join(load_balancer_arn.split(":")[-1].split("/")[-3:])
                    return load_balancer_name

            # Return Unknown if we didn't find a load balancer
            return "Unknown"

        except Exception as e:
            print(f"Error: {e}")
            return "Unknown"


if __name__ == "__main__":
    """Exercise the DashboardMetrics class"""

    # Create the Class and query for metrics
    my_metrics = ECSInfo()
    print(f"Cluster Name: {my_metrics.get_cluster_name()}")
    print(f"Service Name: {my_metrics.get_service_name()}")
    print(f"Load Balancer Name: {my_metrics.get_load_balancer_name()}")



"""CleanData: Example Class that demonstrates data cleanup for Light DataSources using Pandas"""

# Local imports
from workbench.core.transforms.data_to_data.light.data_to_data_light import DataToDataLight
from workbench.utils import pandas_utils


class CleanData(DataToDataLight):
    """CleanData:Class for filtering, sub-setting, and value constraints on Light DataSources

    Common Usage:
        ```python
        clean_data = CleanData(input_data_name, output_data_name)
        clean_data.set_output_tags(["abalone", "clean", "whatever"])
        clean_data.transform(drop_na="any", drop_outliers=True, drop_duplicates=True)
        ```
    """

    def __init__(self, input_data_name: str, output_data_name: str):
        """CleanData Initialization"""

        # Call superclass init
        super().__init__(input_data_name, output_data_name)

    def transform_impl(self, drop_na="any", drop_duplicates=True, drop_outliers=True, **kwargs):
        """Simple Clean Data, will improve later"""

        """
        Notes for later:
        Cleaning data typically involves two phases: Identification and Remediation.

        - Identification
            - Look for NaNs/NULL
            - Make sure all datatypes are correct (df.info()) if not .. fix...
            - Look at distributions/histograms
            - Look for outliers
        - Remediation
            - Drop
            - Fill/Replace
            - Impute the value (using inference/context to fill in a value)
        """

        # Drop Rows that have NaNs in them
        self.output_df = pandas_utils.remove_rows_with_nans(self.input_df, how=drop_na)

        # Drop Duplicates
        if drop_duplicates:
            self.output_df = pandas_utils.drop_duplicates(self.output_df)

        # Drop Outliers
        if drop_outliers:
            self.output_df = pandas_utils.drop_outliers_iqr(self.output_df, scale=2.0)


if __name__ == "__main__":
    """Exercise the CleanData Class"""

    # Create the class with inputs and outputs and invoke the transform
    input_name = "test_data"
    output_name = "test_data_clean"
    data_to_data = CleanData(input_name, output_name)
    data_to_data.set_output_tags(["test", "clean"])
    data_to_data.transform(drop_na="any")



import boto3

# Initialize Boto3 Glue client
glue_client = boto3.client("glue")

# Step 1: Get the existing workflow details
existing_workflow_name = "Workflow"
response = glue_client.get_workflow(Name=existing_workflow_name, IncludeGraph=True)
existing_workflow = response["Workflow"]

# Assume the workflow structure, jobs, and triggers need to be duplicated;
# this example focuses on the workflow duplication and trigger creation.

# Step 2: Create a new workflow (duplicate of the existing one)
new_workflow_name = "Workflow_Test"
glue_client.create_workflow(Name=new_workflow_name, Description="Test Workflow")

# Step 3: Create an 'on-demand' trigger for the new workflow
trigger_name = "on_demand_trigger_for_" + new_workflow_name
glue_client.create_trigger(
    Name=trigger_name,
    WorkflowName=new_workflow_name,
    Type="ON_DEMAND",  # This specifies the trigger is 'on-demand'
    Actions=[
        {
            # Define actions here; typically starting a job or crawler.
            # Example: 'JobName': 'your_glue_job_name'
            # This part depends on what the first step in your workflow is.
        }
    ],
)

print(f"New workflow '{new_workflow_name}' created with an 'on-demand' trigger.")



"""Tests the support for Empty Model Groups"""

from pprint import pprint

# Workbench Imports
from workbench.api.model import Model


model = Model("empty-model-group")


# Test the Model Metrics
def test_list_inference_runs():
    """Test the List Inference Runs"""
    print("\n\n*** List Inference Runs ***")
    pprint(model.list_inference_runs())


def test_performance_metrics():
    """Test the Model Performance Metrics"""
    print("\n\n*** Performance Metrics ***")
    pprint(model.get_inference_metadata())


def test_retrieval_with_capture_name():
    """Test the retrieval of the model metrics using capture Name"""
    capture_list = model.list_inference_runs()
    for capture_name in capture_list:
        print(f"\n\n*** Retrieval with Capture Name ({capture_name}) ***")
        pprint(model.get_inference_metadata(capture_name).head())
        pprint(model.get_inference_metrics(capture_name).head())
        pprint(model.get_inference_predictions(capture_name).head())
        pprint(model.confusion_matrix(capture_name))


def test_validation_predictions():
    print("\n\n*** Validation Predictions ***")
    val_predictions = model._get_validation_predictions()
    if val_predictions is None:
        print(f"Model {model.name} has no validation predictions!")
    else:
        pprint(val_predictions.head())


def test_confusion_matrix():
    print("\n\n*** Confusion Matrix ***")
    pprint(model.confusion_matrix())


def test_metrics_with_capture_name():
    """Test the Performance Metrics using a Capture Name"""
    metrics = model.get_inference_metrics("auto_inference")
    print("\n\n*** Performance Metrics with Capture Name ***")
    pprint(metrics)


if __name__ == "__main__":

    # Set Pandas display options
    import pandas as pd

    pd.set_option("display.max_columns", None)
    pd.set_option("display.max_rows", None)
    pd.set_option("display.width", None)

    # Run the tests
    test_list_inference_runs()
    test_performance_metrics()
    test_retrieval_with_capture_name()
    test_validation_predictions()
    test_confusion_matrix()
    test_metrics_with_capture_name()



"""Helper functions for working with Workbench Lambda Layer"""

import os
import sys
import shutil
import zipfile
import logging

log = logging.getLogger("workbench")


def load_lambda_layer():
    # Path to the directory containing zip files in the Lambda layer
    zip_dir = "/opt/python_zipped"
    extract_path = "/tmp/python_unzipped"

    # Create the extraction path if it doesn't exist
    os.makedirs(extract_path, exist_ok=True)

    # Check for available disk space in the extract path
    min_required_space = 1024 * 1024 * 512  # 512 MB
    total, used, free = shutil.disk_usage(extract_path)
    if free < min_required_space:
        log.error(f"Insufficient disk space. Available: {free / (1024 * 1024):.2f} MB.")
        log.error("Package extraction and installation may fail, if so,  increase ephemeral storage.")
    else:
        log.important(f"Available ephemeral storage should be sufficient: {free / (1024 * 1024):.2f} MB.")

    # Move sklearn from zip_dir to extract_path
    sklearn_dir = os.path.join(zip_dir, "sklearn")
    if os.path.exists(sklearn_dir):
        shutil.move(sklearn_dir, extract_path)
        log.important(f"Moved sklearn from {sklearn_dir} to {extract_path}")

    # Check if the zip directory exists
    if not os.path.exists(zip_dir):
        log.error(f"Zip directory not found: {zip_dir}")
        return

    # Iterate over each file in the zip directory
    for file_name in os.listdir(zip_dir):
        if file_name.endswith(".zip"):
            zip_path = os.path.join(zip_dir, file_name)
            log.important(f"Extracting {zip_path} to {extract_path}")

            # Extract each zip file into the extract path
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(extract_path)

    # Add the extracted path to the System Path (so imports will find it)
    sys.path.append(extract_path)



"""This Script creates Endpoints for Timing Tests

Models:
    - Using abalone-regression to create the two endpoints.
Endpoints:
    - test-timing-serverless, test-timing-realtime
"""

import sys
import logging

# Workbench imports
from workbench.api import Model, Endpoint

log = logging.getLogger("workbench")


if __name__ == "__main__":

    # Grab our Model
    model = Model("abalone-regression")
    if not model.exists():
        log.error("The abalone-regression model does not exist. Please create it first.")
        sys.exit(1)

    # Create the two Endpoints
    if not Endpoint("test-timing-serverless").exists():
        end = model.to_endpoint("test-timing-serverless", tags=["test", "timing"])

    if not Endpoint("test-timing-realtime").exists():
        end = model.to_endpoint("test-timing-realtime", tags=["test", "timing"], serverless=False)



"""Model: Manages AWS Model Package/Group creation and management.

Models are automatically set up and provisioned for deployment into AWS.
Models can be viewed in the AWS Sagemaker interfaces or in the Workbench
Dashboard UI, which provides additional model details and performance metrics
"""

# Workbench Imports
from workbench.core.artifacts.artifact import Artifact
from workbench.core.artifacts.model_core import ModelCore, ModelType  # noqa: F401
from workbench.core.transforms.model_to_endpoint.model_to_endpoint import ModelToEndpoint
from workbench.api.endpoint import Endpoint
from workbench.utils.model_utils import proximity_model, uq_model


class Model(ModelCore):
    """Model: Workbench Model API Class.

    Common Usage:
        ```python
        my_model = Model(name)
        my_model.details()
        my_model.to_endpoint()
        ```
    """

    def details(self, **kwargs) -> dict:
        """Retrieve the Model Details.

        Returns:
            dict: A dictionary of details about the Model
        """
        return super().details(**kwargs)

    def to_endpoint(
        self,
        name: str = None,
        tags: list = None,
        serverless: bool = True,
        mem_size: int = 2048,
        max_concurrency: int = 5,
        instance: str = "ml.t2.medium",
        data_capture: bool = False,
    ) -> Endpoint:
        """Create an Endpoint from the Model.

        Args:
            name (str): Set the name for the endpoint. If not specified, an automatic name will be generated
            tags (list): Set the tags for the endpoint. If not specified automatic tags will be generated.
            serverless (bool): Set the endpoint to be serverless (default: True)
            mem_size (int): The memory size for the Endpoint in MB (default: 2048)
            max_concurrency (int): The maximum concurrency for the Endpoint (default: 5)
            instance (str): The instance type to use for Realtime(serverless=False) Endpoints (default: "ml.t2.medium")
            data_capture (bool): Enable data capture for the Endpoint (default: False)

        Returns:
            Endpoint: The Endpoint created from the Model
        """

        # Ensure the endpoint_name is valid
        if name:
            Artifact.is_name_valid(name, delimiter="-", lower_case=False)

        # If the endpoint_name wasn't given generate it
        else:
            name = self.name.replace("_features", "") + ""
            name = Artifact.generate_valid_name(name, delimiter="-")

        # Create the Endpoint Tags
        tags = [name] if tags is None else tags

        # Create an Endpoint from the Model
        model_to_endpoint = ModelToEndpoint(self.name, name, serverless=serverless, instance=instance)
        model_to_endpoint.set_output_tags(tags)
        model_to_endpoint.transform(
            mem_size=mem_size,
            max_concurrency=max_concurrency,
            data_capture=data_capture,
        )

        # Set the Endpoint Owner and Return the Endpoint
        end = Endpoint(name)
        end.set_owner(self.get_owner())
        return end

    def prox_model(self, prox_model_name: str = None, track_columns: list = None) -> "Model":
        """Create a Proximity Model for this Model

        Args:
            prox_model_name (str, optional): Name of the Proximity Model (if not specified, a name will be generated)
            track_columns (list, optional): List of columns to track in the Proximity Model.

        Returns:
            Model: The Proximity Model
        """
        if prox_model_name is None:
            prox_model_name = self.model_name + "-prox"
        return proximity_model(self, prox_model_name, track_columns=track_columns)

    def uq_model(self, uq_model_name: str = None, train_all_data: bool = False) -> "Model":
        """Create a Uncertainty Quantification Model for this Model

        Args:
            uq_model_name (str, optional): Name of the UQ Model (if not specified, a name will be generated)
            train_all_data (bool, optional): Whether to train the UQ Model on all data (default: False)

        Returns:
            Model: The UQ Model
        """
        if uq_model_name is None:
            uq_model_name = self.model_name + "-uq"
        return uq_model(self, uq_model_name, train_all_data=train_all_data)


if __name__ == "__main__":
    """Exercise the Model Class"""
    from pprint import pprint

    # Retrieve an existing Data Source
    my_model = Model("abalone-regression")
    pprint(my_model.summary())
    pprint(my_model.details())

    # Create an Endpoint from the Model
    my_endpoint = my_model.to_endpoint()
    pprint(my_endpoint.summary())



import time
from workbench.api import Meta, Endpoint

meta = Meta()


# Check if artifact has sageworks metadata
def has_sageworks_meta(meta: dict) -> bool:
    return any("sageworks" in key for key in meta)


# Loop through the Models and update the metadata
for name in meta.endpoints()["Name"].values:
    end = Endpoint(name)

    # Loop over the metadata and migrate
    meta = end.workbench_meta()

    # Check if artifact has sageworks metadata
    if not has_sageworks_meta(meta):
        print(f"Skipping {name} as it does not have sageworks metadata")
        continue

    # Build new metadata
    print(f"Starting migration for {name}...")
    new_meta = {}
    for key, value in meta.items():
        if "sageworks" in key:
            print(f"Replacing {key} with {key.replace('sageworks', 'workbench')}")
            new_meta[key.replace("sageworks", "workbench")] = value

    # Now Delete all the old keys
    for key in meta.keys():
        if "sageworks" in key:
            end.delete_metadata(key)

    # Add new meta
    end.upsert_workbench_meta(new_meta)
    print(f"Migrating done for {name}...")
    time.sleep(5)



"""SQL based Value Counts: Compute Value Counts for all columns in a DataSource using SQL"""

import logging
import pandas as pd

# Workbench Imports
from workbench.core.artifacts.data_source_abstract import DataSourceAbstract

# Workbench Logger
log = logging.getLogger("workbench")


def value_counts(data_source: DataSourceAbstract) -> dict[dict]:
    """Compute 'value_counts' for all the string columns in a DataSource
    Args:
        data_source: The DataSource that we're computing value_counts on
    Returns:
        dict(dict): A dictionary of value counts for each column in the form
             {'col1': {'value_1': X, 'value_2': Y, 'value_3': Z,...},
              'col2': ...}
    """

    # Grab the DataSource computation table name
    table = data_source.view("computation").table

    # For every column in the table that is string, compute the value_counts
    data_source.log.info("Computing value_counts for all string columns...")
    value_count_dict = dict()
    column_details = data_source.view("computation").column_details()
    for column, data_type in column_details.items():
        if data_type in ["string", "boolean"]:
            # Combined query to get both top and bottom counts
            query = (
                f'(SELECT "{column}", count(*) as workbench_count '
                f'FROM "{table}" '
                f'GROUP BY "{column}" ORDER BY workbench_count DESC LIMIT 20) '
                f"UNION ALL "
                f'(SELECT "{column}", count(*) as workbench_count '
                f'FROM "{table}" '
                f'GROUP BY "{column}" ORDER BY workbench_count ASC LIMIT 20)'
            )
            log.debug(query)
            result_df = data_source.query(query)

            # Convert Int64 (nullable) to int32 so that we can serialize to JSON
            result_df["workbench_count"] = result_df["workbench_count"].astype("int32")

            # If the column is boolean, convert the values to True/False strings (for JSON)
            if result_df[column].dtype == "boolean":
                result_df[column] = result_df[column].astype("string")

            # Convert any NA values to 'NaN' so that we can serialize to JSON
            result_df.fillna("NaN", inplace=True)

            # If all of our counts equal 1 we can drop most of them
            if result_df["workbench_count"].sum() == result_df.shape[0]:
                result_df = result_df.iloc[:5]

            # Convert the result_df into a dictionary
            value_count_dict[column] = dict(zip(result_df[column], result_df["workbench_count"]))

    # Return the value_count data
    return value_count_dict


if __name__ == "__main__":
    """Exercise the SQL Value Counts Functionality"""
    from pprint import pprint
    from workbench.api.data_source import DataSource

    # Setup Pandas output options
    pd.set_option("display.max_colwidth", 50)
    pd.set_option("display.max_columns", 15)
    pd.set_option("display.width", 1000)

    # Retrieve a Data Source
    my_data = DataSource("abalone_data")

    # Verify that the Athena Data Source exists
    assert my_data.exists()

    # What's my Workbench Name
    print(f"Name: {my_data.name}")

    # Get sample rows for this DataSource
    my_value_counts = value_counts(my_data)
    print("\nValue Counts")
    pprint(my_value_counts)



"""This Script Deletes the Workbench Artifacts used for stress tests"""

import time
from workbench.api import DataSource, FeatureSet, Model, Endpoint


if __name__ == "__main__":

    # Generated Model Names
    model_names = [f"stress-model-{i}" for i in range(16)]
    datasource_names = [f"{model.replace('-model-', '_data_')}" for model in model_names]
    feature_names = [f"{model.replace('-model-', '_features_')}" for model in model_names]
    endpoint_names = [f"{model}" for model in model_names]

    # Delete the DataSources
    for datasource_name in datasource_names:
        DataSource.managed_delete(datasource_name)

    # Delete the FeatureSets
    for feature_name in feature_names:
        FeatureSet.managed_delete(feature_name)

    # Delete the Models
    for model_name in model_names:
        Model.managed_delete(model_name)

    # Delete the Endpoints
    for endpoint_name in endpoint_names:
        Endpoint.managed_delete(endpoint_name)

    time.sleep(5)
    print("All stress test artifacts should now be deleted!")



"""TargetGradients: Compute Feature Space to Target Gradients"""

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.neighbors import KNeighborsRegressor
from sklearn.preprocessing import StandardScaler

# Workbench Imports
from workbench.utils.pandas_utils import remove_rows_with_nans


class TargetGradients(BaseEstimator, TransformerMixin):
    """
    A custom transformer for computing gradients between observations that are close in feature space.
    This transformer adds 'feature_diff', 'target_diff' and 'target_gradient' columns to the DataFrame.

    Attributes:
        n_neighbors (int): Number of neighbors to consider.
    """

    def __init__(self, n_neighbors: int = 2):
        """Initialize the TargetGradients with the specified parameters.

        Args:
            n_neighbors (int): Number of neighbors to consider (default: 2)
        """
        self.n_neighbors = n_neighbors
        self.scalar = StandardScaler()
        self.knn = KNeighborsRegressor(n_neighbors=self.n_neighbors, algorithm="ball_tree", metric="euclidean")
        self.X = None
        self.y = None
        super().__init__()

    def fit(self, X: pd.DataFrame, y: pd.Series) -> BaseEstimator:
        """
        Standardize the features and fit the internal KNN model

        Args:
            X (pd.DataFrame): The input features.
            y (pd.Series, optional): The target variable.

        Returns:
            self: Returns an instance of self.
        """

        # Drop rows with NaNs (we need to combine X and y, so that rows with NaNs are dropped together)
        df = pd.concat([X, y], axis=1)
        df = remove_rows_with_nans(df)
        X = df.drop(columns=[y.name])
        y = df[y.name]

        # Store the input data
        self.X = X
        self.y = y

        # Standardize the features
        X_norm = self.scalar.fit_transform(X)

        # Fit the KNN model
        self.knn.fit(X_norm, self.y)

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms the input DataFrame by adding 'feature_diff', 'target_diff' and 'target_gradient'  columns.

        Args:
            X (pd.DataFrame): The input features.

        Returns:
            pd.DataFrame: The transformed DataFrame with additional columns.
        """

        # Drop rows with NaNs in the DataFrame
        X = remove_rows_with_nans(X)

        # Standardize the features
        df_norm = pd.DataFrame(self.scalar.fit_transform(X), columns=X.columns, index=X.index)

        # Initialize arrays to accumulate results
        feature_diff = []
        target_diff = []
        target_gradient = []

        # Compute the gradients
        for index, row in df_norm.iterrows():
            # Find the nearest neighbors
            distances, indices = self.knn.kneighbors([row], n_neighbors=self.n_neighbors)

            # Find the index of the nearest neighbor that is not the observation itself
            nn_index = indices[0][1] if indices[0][0] == index else indices[0][0]

            # Compute the difference in feature space and target space
            diff_feature = distances[0][1] if indices[0][0] == index else distances[0][0]
            diff_target = abs(self.y.iloc[index] - self.y.iloc[nn_index])

            # Calculate the target gradient, handling division by zero
            gradient_target = float("inf") if diff_feature == 0 else diff_target / diff_feature

            # Accumulate results
            feature_diff.append(diff_feature)
            target_diff.append(diff_target)
            target_gradient.append(gradient_target)

        # Add the accumulated results to the DataFrame
        df_norm["feature_diff"] = feature_diff
        df_norm["target_diff"] = target_diff
        df_norm["target_gradient"] = target_gradient

        return df_norm

    def fit_transform(self, X: pd.DataFrame, y: pd.Series, **fit_params):
        """
        Fits the model and transforms the input DataFrame by adding 'feature_diff',
        'target_diff' and 'target_gradient'  columns.

        Args:
            X (pd.DataFrame): The input features.
            y (pd.Series, optional): The target variable.
        """
        self.fit(X, y)
        return self.transform(X)


if __name__ == "__main__":
    """Example usage of the TargetGradients class"""
    from workbench.api.feature_set import FeatureSet
    from workbench.api.model import Model

    # Load the Abalone FeatureSet
    fs = FeatureSet("abalone_features")
    df = fs.pull_dataframe()

    # Grab the target and feature columns from the model
    model = Model("abalone-regression")
    target_column = model.target()
    feature_columns = model.features()

    # Initialize the TargetGradients
    target_gradients = TargetGradients(n_neighbors=2)
    result_df = target_gradients.fit_transform(df[feature_columns], df[target_column])

    # Print the result DataFrame
    print(result_df)



"""Tests for the FeatureSet functionality"""

# Workbench Imports
from workbench.core.artifacts.feature_set_core import FeatureSetCore


def test():
    """Simple test of the FeatureSet functionality"""
    from pprint import pprint

    # Grab a FeatureSet object and pull some information from it
    my_features = FeatureSetCore("test_features")

    # Call the various methods

    # Let's do a check/validation of the feature set
    print(f"Feature Set Check: {my_features.exists()}")

    # How many rows and columns?
    num_rows = my_features.num_rows()
    num_columns = my_features.num_columns()
    print(f"Rows: {num_rows} Columns: {num_columns}")

    # What are the column names?
    print(my_features.columns)

    # Get Tags associated with this Feature Set
    print(f"Tags: {my_features.get_tags()}")

    # Get ALL the AWS Metadata associated with this Feature Set
    print("\n\nALL Meta")
    pprint(my_features.aws_meta())

    # Now delete the AWS artifacts associated with this Feature Set
    # print('Deleting Workbench Feature Set...')
    # my_features.delete()


if __name__ == "__main__":
    test()
